# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for tracking and monitoring memory and resource usage."""
# Python standard Libraries
# Python standard Libraries
import logging
import socket
import os
import resource
import threading
from ast import literal_eval
from sys import argv
from datetime import datetime
import platform

# Installed libraires
import pandas as pd

# Geoips libraries
from geoips.filenames.base_paths import PATHS as gpaths

from geoips.utils.context_managers import import_optional_dependencies

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import psutil


def print_mem_usage(logstr="", verbose=False):
    """Print memory usage to LOG.info.

    * By default include psutil output.
    * If verbose is True, include output from both psutil and resource packages.
    """
    # If psutil / socket / resource are not imported, do not fail
    usage_dict = {}
    try:
        vmem_percent = psutil.virtual_memory().percent
        LOG.info(
            "virtual perc: %s on %s %s",
            str(vmem_percent),
            str(socket.gethostname()),
            logstr,
        )
        swap_percent = psutil.swap_memory().percent
        LOG.info(
            "swap perc:    %s on %s %s",
            str(swap_percent),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        vmem_percent = swap_percent = "nan"
        LOG.info(
            "%s: psutil or socket not defined, no percent memusg output", str(resp)
        )
    try:
        highest = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        LOG.info(
            "highest:      %s on %s %s",
            str(highest),
            str(socket.gethostname()),
            logstr,
        )
    except NameError as resp:
        highest = "nan"
        LOG.info(
            "%s: resource or socket not defined, no highest memusg output", str(resp)
        )

    if verbose:
        usage_dict = print_resource_usage(logstr)
    usage_dict["memusg_virtual"] = vmem_percent
    usage_dict["memusg_swap"] = swap_percent
    usage_dict["memusg_highest"] = highest
    return usage_dict


def print_resource_usage(logstr=""):
    """Print verbose resource usage, using "resource" package."""
    usage_dict = {}
    try:
        usage = resource.getrusage(resource.RUSAGE_SELF)
        for name, desc in [
            ("ru_utime", "RESOURCE " + logstr + " User time"),
            ("ru_stime", "RESOURCE " + logstr + " System time"),
            ("ru_maxrss", "RESOURCE " + logstr + " Max. Resident Set Size"),
            ("ru_ixrss", "RESOURCE " + logstr + " Shared Memory Size"),
            ("ru_idrss", "RESOURCE " + logstr + " Unshared Memory Size"),
            ("ru_isrss", "RESOURCE " + logstr + " Stack Size"),
            ("ru_inblock", "RESOURCE " + logstr + " Block inputs"),
            ("ru_oublock", "RESOURCE " + logstr + " Block outputs"),
        ]:
            LOG.info("%-25s (%-10s) = %s", desc, name, getattr(usage, name))
            usage_dict[name] = getattr(usage, name)
    except NameError:
        LOG.info("resource not defined")
    return usage_dict


class PidLog:
    """Track a PID and all children.

    * Requires psutil and threading
    """

    def __init__(self, inpid, logstr=""):

        self.ps_parent = psutil.Process(pid=inpid)
        self.own_pid = os.getpid()
        self.logstr = logstr
        self.pid_bool = self.ps_parent.is_running()
        # allows for other process to run
        self.track_pid_thread = threading.Thread(
            target=self.track_pids, name="TrackPID", daemon=True
        )
        self.track_pid_thread.start()

    def track_pids(self):
        """Track pids and create a dict of values."""
        self.usage_dict = {
            "cpu_percent": [],
            "thread_count": [],
            "unique_set_size": [],
            "res_set_size": [],
            "utc_datetime": [],
        }
        if platform.system() == "Linux":
            self.usage_dict["cpu_count"] = []

        while self.pid_bool:

            ps_child = self.ps_parent.children(recursive=True)

            # filter out own pid and any children spawned
            all_pids = [self.ps_parent] + [
                i
                for i in ps_child
                for j in i.parents()
                if i.pid != self.own_pid and j.pid != self.own_pid
            ]

            tmp_cpu, cpu_per, thrd_cnt, uss_tmp, rss_tmp = [], 0, 0, 0, 0

            for i in all_pids:
                try:
                    # CPU usage with oneshot doesn't work

                    with i.oneshot():
                        # cpu_num only works for Linux, FreeBSD, SunOS
                        if platform.system() == "Linux":
                            tmp_cpu += [i.cpu_num()]
                        thrd_cnt += i.num_threads()
                        tmp_mem = i.memory_full_info()
                        uss_tmp += tmp_mem.uss
                        rss_tmp += tmp_mem.rss
                    # might want average over all cpus?
                    cpu_per += i.cpu_percent(interval=0.1)
                except psutil.Error:
                    # pid expired quickly, need to remove invalid values
                    continue

            dtime = datetime.utcnow().strftime("%Y%m%d_%H%M%S_%f")
            self.usage_dict["cpu_percent"].append(cpu_per)
            if platform.system() == "Linux":
                self.usage_dict["cpu_count"].append(len(set(tmp_cpu)))
            self.usage_dict["thread_count"].append(thrd_cnt)
            self.usage_dict["unique_set_size"].append(uss_tmp)
            self.usage_dict["res_set_size"].append(rss_tmp)
            self.usage_dict["utc_datetime"].append(dtime)

        return

    def print_mem_usg(self, logstr="", verbose=False):
        """Print verbose resouce usage."""
        usage_dict = {}
        vmem_percent = psutil.virtual_memory().percent
        LOG.info(
            "virtual perc: %s on %s %s",
            str(vmem_percent),
            str(socket.gethostname()),
            self.logstr,
        )
        #

        swap_percent = psutil.swap_memory().percent
        LOG.info(
            "swap perc:    %s on %s %s",
            str(swap_percent),
            str(socket.gethostname()),
            self.logstr,
        )

        if len(self.usage_dict["unique_set_size"]) > 0:
            ram_usg = max(self.usage_dict["unique_set_size"])
            highest = max(self.usage_dict["res_set_size"])
        else:
            ram_usg = 0
            highest = 0

        LOG.info(
            "highest ram:    %s on %s %s",
            str(ram_usg),
            str(socket.gethostname()),
            self.logstr,
        )

        LOG.info(
            "highest rss:      %s on %s %s",
            str(highest),
            str(socket.gethostname()),
            self.logstr,
        )

        if verbose:
            usage_dict = self.print_resource_usage()

        usage_dict["memusg_virtual"] = vmem_percent
        usage_dict["memusg_swap"] = swap_percent
        usage_dict["memusg_highest"] = highest

        return usage_dict

    def print_resource_usage(self):
        """Print verbose resource usage, using "resource" package."""
        usage_dict = {}
        try:
            usage = resource.getrusage(resource.RUSAGE_SELF)
            for name, desc in [
                ("ru_utime", "RESOURCE " + self.logstr + " User time"),
                ("ru_stime", "RESOURCE " + self.logstr + " System time"),
                ("ru_maxrss", "RESOURCE " + self.logstr + " Max. Resident Set Size"),
                ("ru_ixrss", "RESOURCE " + self.logstr + " Shared Memory Size"),
                ("ru_idrss", "RESOURCE " + self.logstr + " Unshared Memory Size"),
                ("ru_isrss", "RESOURCE " + self.logstr + " Stack Size"),
                ("ru_inblock", "RESOURCE " + self.logstr + " Block inputs"),
                ("ru_oublock", "RESOURCE " + self.logstr + " Block outputs"),
            ]:
                LOG.info("%-25s (%-10s) = %s", desc, name, getattr(usage, name))
                usage_dict[name] = getattr(usage, name)
        except NameError:
            LOG.info("resource not defined")
        usage_dict["ru_cpuusg"] = max(self.usage_dict["cpu_percent"])  # [-1]
        usage_dict["ru_threads"] = max(self.usage_dict["thread_count"])
        if platform.system() == "Linux":
            usage_dict["ru_cpucnt"] = max(self.usage_dict["cpu_count"])
        usage_dict["ru_uss"] = self.usage_dict["unique_set_size"][-1]
        usage_dict["ru_maxuss"] = max(self.usage_dict["unique_set_size"])
        LOG.info(
            "RESOURCE "
            + self.logstr
            + " Max. User Set Size (RAM) = "
            + str(max(self.usage_dict["unique_set_size"]))
        )
        LOG.info(
            "RESOURCE "
            + self.logstr
            + " Max. CPU % Usage = "
            + str(max(self.usage_dict["cpu_percent"]))
        )
        return usage_dict

    def save_exit(self):
        """Exit the thread cleanly."""
        LOG.info("Stopping resource tracking.")
        self.pid_bool = False
        self.track_pid_thread.join()

    def save_csv(self):
        """Save a csv file to output."""
        usg_dict = {
            "Time [UTC]": self.usage_dict["utc_datetime"],
            "CPU Percent": self.usage_dict["cpu_percent"],
            "Thread Count": self.usage_dict["thread_count"],
            "USS [RAM bytes]": self.usage_dict["unique_set_size"],
            "RSS [bytes]": self.usage_dict["res_set_size"],
        }
        if platform.system() == "Linux":
            usg_dict["CPU Count"] = self.usage_dict["cpu_count"]
        df = pd.DataFrame(usg_dict)

        outdir = os.path.join(gpaths["GEOIPS_OUTDIRS"], "memory_logs")
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        outname = "geoips_memtrack_p{}.csv".format(self.ps_parent.pid)
        outpath = outdir + "/" + outname
        LOG.info("Saving csv memory track to {}".format(outpath))

        df.to_csv(outpath, index=False)


def single_track_pid(procpid):
    """Output a snapshot of a pid usage on server.

    * Requires an input pid.
    """
    ps_parent = psutil.Process(pid=procpid)
    own_pid = os.getpid()
    ps_child = ps_parent.children(recursive=True)

    # filter out own pid and any children spawned
    all_pids = [ps_parent] + [
        i
        for i in ps_child
        for j in i.parents()
        if i.pid != own_pid and j.pid != own_pid
    ]

    tmp_cpu, cpu_per, thrd_cnt, uss_tmp, rss_tmp = [], 0, 0, 0, 0

    for i in all_pids:
        try:
            # CPU usage with oneshot doesn't work

            with i.oneshot():
                # cpu_num only works for Linux, FreeBSD, SunOS
                if platform.system() == "Linux":
                    tmp_cpu += [i.cpu_num()]
                thrd_cnt += i.num_threads()
                tmp_mem = i.memory_full_info()
                uss_tmp += tmp_mem.uss
                rss_tmp += tmp_mem.rss
            # might want average over all cpus?
            cpu_per += i.cpu_percent(interval=0.1)
        except psutil.Error:
            # pid expired quickly, need to remove invalid values
            continue

    if platform.system() == "Linux":
        # outputs CPU percent, CPU count, Thread count, USS, RSS
        print(
            "{}, {}, {}, {}, {}".format(
                cpu_per, len(set(tmp_cpu)), thrd_cnt, uss_tmp, rss_tmp
            )
        )
    else:
        # outputs CPU percent, Thread count, USS, RSS
        print("{}, {}, {}, {}".format(cpu_per, thrd_cnt, uss_tmp, rss_tmp))


if __name__ == "__main__":
    """Track a PID given from command line."""
    pid_track = single_track_pid(literal_eval(argv[1]))



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Algorithms interface module."""

from geoips.interfaces.base import BaseModuleInterface


class AlgorithmsInterface(BaseModuleInterface):
    """Data manipulations to apply to the dataset."""

    name = "algorithms"

    required_args = {
        "scalar_to_scalar": [],
        "single_channel": ["arrays"],
        "channel_combination": ["arrays"],
        "list_numpy_to_numpy": ["arrays"],
        "xarray_to_numpy": ["xobj"],
        "xarray_to_xarray": ["xobj", "variables", "product_name"],
        "rgb": ["arrays"],
        "xarray_dict_to_xarray": ["xarray_dict"],
        "xarray_dict_dict_to_xarray": ["xarray_dict_dict"],
        "xarray_dict_to_xarray_dict": ["xarray_dict"],
        "xarray_dict_area_def_to_numpy": ["xarray_dict", "area_def"],
    }

    required_kwargs = {
        "scalar_to_scalar": ["value"],
        "single_channel": [],
        "channel_combination": [],
        "xarray_to_numpy": [],
        "xarray_to_xarray": [],
        "list_numpy_to_numpy": [],
        "rgb": [],
        "xarray_dict_to_xarray": [],
        "xarray_dict_dict_to_xarray": [],
        "xarray_dict_to_xarray_dict": [],
        "xarray_dict_area_def_to_numpy": [],
    }


algorithms = AlgorithmsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Night_Vis_IR" product.

This algorithm expects two VIIRS channels (DNBRad and M16BT) for a RGB image
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "RGB_Default"


def call(arrays):
    """Apply default RGB algorithm.

    Plots RGB with the red gun being the first variable specified in the
    product YAML, green second, and blue third.

    Note this is currently entirely unused, and is included here for
    reference/completeness.  Eventually we may want to fully support
    this generalized RGB default algorithm, where you can pass in
    ranges, channel combinations, etc for RGB output, but for now it
    will ONLY plot arrays 0, 1, 2 as red, green, blue respectively,
    with no adjustments/combinations.

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    red = arrays[0]  # Red gun: First listed variable
    grn = arrays[1]  # Green gun: Second listed variable
    blu = arrays[2]  # Blue gun: Third listed variable

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



"""Unit test module for the CLI command 'geoips list registries'."""

import pytest

from geoips.commandline.commandline_interface import main
from tests.unit_tests.commandline.cli_top_level_tester import (
    generate_id,
    check_valid_command_using_monkeypatch,
)

valid_args = [
    ["geoips", "list", "registries"],
    ["geoips", "list", "registries", "-c", "package", "json"],
    ["geoips", "list", "registries", "-p", "geoips"],
    ["geoips", "list", "registries", "-n", "geoips.plugin_packages"],
    ["geoips", "list", "registries", "-r"],
    ["geoips", "list", "registries", "-n", "geoips.plugin_packages", "-r"],
]
valid_expected = [
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "all",
        "long": True,
        "columns": None,
        "namespace": "geoips.plugin_packages",
        "relpath": False,
    },
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "all",
        "long": True,
        "columns": ["package", "json"],
        "namespace": "geoips.plugin_packages",
        "relpath": False,
    },
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "geoips",
        "long": True,
        "columns": None,
        "namespace": "geoips.plugin_packages",
        "relpath": False,
    },
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "all",
        "long": True,
        "columns": None,
        "namespace": "geoips.plugin_packages",
        "relpath": False,
    },
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "all",
        "long": True,
        "columns": None,
        "namespace": "geoips.plugin_packages",
        "relpath": True,
    },
    {
        "log_level": "interactive",
        "command": "geoips list registries",
        "package_name": "all",
        "long": True,
        "columns": None,
        "namespace": "geoips.plugin_packages",
        "relpath": True,
    },
]
invalid_args = [
    ["geoips", "list", "registries", "-n", "fake.plugin_packages"],
    ["geoips", "list", "registries", "-c", "not_a_column"],
    ["geoips", "list", "registries", "-p", "fake_package"],
]


@pytest.mark.parametrize("cli_args", [args for args in invalid_args], ids=generate_id)
def test_list_registries_invalid(monkeypatch, cli_args):
    """Test the CLI command 'geoips list registries' using invalid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    """
    monkeypatch.setattr("sys.argv", cli_args)
    with pytest.raises(SystemExit):
        main()


@pytest.mark.parametrize(
    "cli_args,expected",
    [(args, expected) for args, expected in zip(valid_args, valid_expected)],
    ids=generate_id,
)
def test_list_registries_valid(monkeypatch, capsys, cli_args, expected):
    """Test the CLI command 'geoips list registries' using valid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    expected: dict
        - A dictionary of expected argument values that are returned from the command
          being ran from args.
    """
    check_valid_command_using_monkeypatch(monkeypatch, cli_args, expected)
    captured = capsys.readouterr()
    headers = {
        "package": "GeoIPS Package",
        "json": "JSON Path",
        "yaml": "YAML Path",
    }
    if "-c" in cli_args:
        selected = cli_args[cli_args.index("-c") + 1 :]
    else:
        selected = list(headers.keys())
    for col in selected:
        assert headers[col] in captured.out



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Overpass predictor, based on Two Line Element files."""

# Standard libraries
from datetime import timedelta
import math
from numpy import linspace

# Installed libraries
import ephem
from pyresample import spherical_geometry

from logging import getLogger

LOG = getLogger()


def check_tle_name_to_passed_names(tle_name, satellite_names_list):
    """Check if the satellite name in the TLE files is in the satellite names list.

    Satellite names in the TLE files may be longer than the names passed to the
    overpass predictor For example, a user might request 'GCOM-W1', and the name
    in the TLE file is 'GCOM-W1 (SHIZUKU)'.

    Parameters
    ----------
    tle_name : str
        satellite name read from the TLE file
    satellite_names_list : list of str
        list of user specified satellites to read from TLE file

    Returns
    -------
    bool
        True if tle_name is in the passed satellite names list
    """
    satellite_in_list = any(
        [x.lower() in tle_name.lower() for x in satellite_names_list]
    )
    return satellite_in_list


def read_satellite_tle(tlefile, satellite_list):
    """Open and extract satellite infromation from TLE file.

    Parameters
    ----------
    tlefile : str
        file path of TLE
    satellite_list : list of str
        list of satellites to read from TLE

    Returns
    -------
    dict
        satellite TLE data
    """
    with open(tlefile, "r") as tfile:
        tle_content = tfile.readlines()
    satellite_tles = {}
    for i, line in enumerate(tle_content):
        if line[0] in ["1", "2"]:
            continue
        satellite = line.strip()
        if check_tle_name_to_passed_names(satellite, satellite_list):
            line1 = tle_content[i + 1].strip()
            line2 = tle_content[i + 2].strip()
            clean_name = satellite.split(" (")[0]
            satellite_tles[clean_name] = {"line1": line1, "line2": line2}
    return satellite_tles


def floor_minute(datetime_obj):
    """Remove seconds and microseconds from datetime object.

    Parameters
    ----------
    datetime_obj : datetime.datetime
        datetime

    Returns
    -------
    datetime.datetime
        datetime with no seconds/microseconds
    """
    second = datetime_obj.second
    micro = datetime_obj.microsecond
    return datetime_obj - timedelta(seconds=second, microseconds=micro)


def calculate_overpass(tle, observer_lat, observer_lon, date, satellite_name):
    """Calculate next overpass for a satellite at an observer location and time.

    Parameters
    ----------
    tle : ephem.EarthSatellite
        tle for satellite
    observer_lat : float
        observer latitude
    observer_lon : float
        observer longitude
    date : datetime.datetime
        start time for next overpass
    satellite_name : str
        name of satellite

    Returns
    -------
    dict
        next overpass information
    """
    sector = ephem.Observer()
    sector.lon = str(observer_lon)
    sector.lat = str(observer_lat)
    sector.date = date.strftime("%Y/%m/%d %H:%M:%S")
    sector.elevation = 0.0
    moon = ephem.Moon()
    sun = ephem.Sun()
    moon.compute(sector)
    sun.compute(sector)
    if not sun.rise_time:
        LOG.info(
            "%s:Something went wrong when calculating sun rise time for %s!",
            satellite_name,
            date,
        )
        return None
    if not sun.set_time:
        LOG.info(
            "%s:Something went wrong when calculating sun set time for %s!",
            satellite_name,
            date,
        )
        return None
    sunrise = sun.rise_time.datetime()
    sunset = sun.set_time.datetime()
    if sunset <= sunrise:
        sunset += timedelta(days=1)
    opass_info = {
        "check datetime": date,
        "sunrise": sunrise,
        "sunset": sunset,
        "moon phase": moon.phase,
    }
    try:
        opass = sector.next_pass(tle)
        is_geostationary = False
        max_alt_time = floor_minute(opass[2].datetime())
        opass_info["rise time"] = floor_minute(opass[0].datetime())
        opass_info["max altitude time"] = max_alt_time
        opass_info["max altitude"] = opass[3]
        opass_info["set time"] = floor_minute(opass[4].datetime())
        opass_info["is daytime"] = (max_alt_time >= sunrise) & (max_alt_time < sunset)
        opass_info["is geostationary"] = False
    except ValueError as resp:
        if "circumpolar" in str(resp):
            is_geostationary = True
            is_above_horizon = True
        else:
            # LOG.info(resp)
            return None
        opass_info["rise time"] = date
        opass_info["max altitude time"] = date
        opass_info["max altitude"] = date
        opass_info["set time"] = date
        opass_info["is geostationary"] = is_geostationary
        opass_info["above horizon"] = is_above_horizon
        opass_info["is daytime"] = (date >= sunrise) & (date < sunset)
    except AttributeError as resp:
        LOG.info(
            "%s: Something when wrong with calculating the next overpass for %s "
            "(AttributeError)",
            satellite_name,
            date,
        )
        LOG.debug(resp)
        return None
    except TypeError as resp:
        LOG.info(
            "%s: Something when wrong with calculating the next overpass for %s "
            "(TypeError)",
            satellite_name,
            date,
        )
        LOG.debug(resp)
        return None
    if is_geostationary:
        tle.compute(sector)
    else:
        tle.compute(opass_info["max altitude time"])
    opass_info["sublon"] = math.degrees(tle.sublong)
    opass_info["sublat"] = math.degrees(tle.sublat)
    ptsatdeg = spherical_geometry.Coordinate(
        lon=opass_info["sublon"], lat=opass_info["sublat"]
    )
    ptsecdeg = spherical_geometry.Coordinate(
        lon=math.degrees(sector.lon), lat=math.degrees(sector.lat)
    )
    earth_radius_km = 6730.0
    cpa_km = ptsatdeg.distance(ptsecdeg) * earth_radius_km
    opass_info["closest pass approach (km)"] = cpa_km
    return opass_info


def create_sector_observers(ll_lon, ll_lat, ur_lon, ur_lat, spacing=50):
    """Create multiple prediction points for large sectors.

    Parameters
    ----------
    ll_lon : float
        Lower left longitude of sector.
    ll_lat : float
        Lower left latitude of sector.
    ur_lon : float
        Upper right longtitude of sector.
    ur_lat : float
        Upper right latitude of sector.
    spacing : int, optional
        Spacing between latitude/longitude observer points (degrees), by default 50

    Returns
    -------
    list
        List of observer points for sector.
    """
    if ur_lon < ll_lon:
        # Check if sector crosses dateline.
        # Convert lons to 0 - 360 just to make things easier.
        ur_lon = 360 + ur_lon
    # Find how wide and tall the sector is
    lon_delta = abs((ur_lon - spacing) - (ll_lon - spacing))
    lat_delta = abs((ur_lat - spacing) - (ll_lat - spacing))
    # Determine how many observers we'd expect for this sector.
    n_lat_points = round(round(lat_delta) / spacing)
    n_lon_points = round(round(lon_delta) / spacing)
    # Create list of latitude observer points.
    # Only try to create multiple observers if the expected
    # number of observers is greater than one.
    if n_lat_points > 1:
        check_lats = linspace(ll_lat + spacing / 2, ur_lat - spacing / 2, n_lat_points)
    else:
        check_lats = [(ur_lat + ll_lat) / 2]
    # Create list of longitude observer points.
    # Only try to create multiple observers if the expected
    # number of observers is greater than one.
    if n_lon_points > 1:
        check_lons = linspace(ll_lon + spacing / 2, ur_lon - spacing / 2, n_lon_points)
        # Convert back to -180 - 180
        check_lons[check_lons > 180] -= 360
    else:
        center_lon = (ur_lon + ll_lon) / 2
        if center_lon > 180:
            # Convert lons back to -180 - 180
            center_lon -= 360
        check_lons = [center_lon]
    check_points = []
    for clat in check_lats:
        for clon in check_lons:
            check_points.append([clat, clon])
    return check_points


def predict_satellite_overpass(
    tlefile,
    satellite_name,
    satellite_tle,
    area_def,
    start_datetime,
    observer_spacing=50,
):
    """Estimate next satellite overpass information with ephem.

    Parameters
    ----------
    tlefile : str
        file path of TLE
    satellite_name : str
        name of satellite
    satellite_tle : dict
        dictionary holding satellite tle line1 and line2 data
    area_def : pyresample AreaDefinition
        area definition
    start_datetime :datetime.datetime
        start time to find the next available overpass
    observer_spacing : float
        Spacing (degrees) between observer points in sector. If domain exceeds the
        specified spacing, multiple observer are automatically added across the sector.

    Returns
    -------
    dict
        dictionary holding next overpass information
    """
    tle = ephem.readtle(tlefile, satellite_tle["line1"], satellite_tle["line2"])
    ll_lon, ll_lat, ur_lon, ur_lat = area_def.area_extent_ll
    observers = create_sector_observers(
        ll_lon, ll_lat, ur_lon, ur_lat, spacing=observer_spacing
    )
    total_observers = len(observers)
    overpasses = {}
    LOG.info(
        "Running %s overpass predictor for %s. Total observers: %s",
        satellite_name,
        area_def.area_id,
        total_observers,
    )
    for i, observer in enumerate(observers):
        observer_lat, observer_lon = observer
        opass_info = calculate_overpass(
            tle, observer_lat, observer_lon, start_datetime, satellite_name
        )
        if isinstance(opass_info, type(None)):
            # Either something went wrong in the predictor, or found a
            # geostationary satellite that does not overpass the sector, ever!
            continue
        opass_key = "_".join(
            [
                satellite_name,
                area_def.area_id,
                opass_info["max altitude time"].strftime("%Y%m%dT%H%MZ"),
            ]
        )
        if opass_key not in overpasses:
            overpasses[opass_key] = opass_info
    return overpasses


def predict_overpass_area_def(
    tlefile, area_definition, satellite_list, start_datetime, observer_spacing=50
):
    """Predict satellite overpass for an area_definition.

    Parameters
    ----------
    tlefile : str
        file path of TLE
    area_definition : pyresample AreaDefinition
        pyresample area definition
    satellite_list : list
        list of satellites to predict the overpass times
    start_datetime : datetime.datetime
        start time to find the next available overpass
    observer_spacing : float
        Spacing (degrees) between observer points in sector. If domain exceeds the
        specified spacing, multiple observer are automatically added across the sector.

    Returns
    -------
    dict
        dictionary holding next satellite overpass estimates
        (sorted by satellite -> overpass info)
    """
    if not isinstance(satellite_list, list):
        raise TypeError("satellite_list must be type list")
    area_def_overpasses = {}
    satellite_tle_dict = read_satellite_tle(tlefile, satellite_list)
    for satellite, sat_tle in satellite_tle_dict.items():
        next_overpass = predict_satellite_overpass(
            tlefile,
            satellite,
            sat_tle,
            area_definition,
            start_datetime,
            observer_spacing=observer_spacing,
        )
        if next_overpass:
            area_def_overpasses[satellite] = next_overpass
    return area_def_overpasses


def predict_overpass_yaml(
    tlefile,
    sectorfile,
    sector_list,
    satellite_list,
    start_datetime,
    observer_spacing=50,
):
    """Predict satellite overpass for sectors from a given yaml sector file.

    Parameters
    ----------
    tlefile : str
        file path of TLE
    sectorfile  : str
        file path of sectorfile
    sector_list : list
        list of sectors held within the sectorfile
    satellite_list : list
        list of satellites to predict the overpass times
    start_datetime : datetime.datetime
        start time to find the next available overpass
    observer_spacing : float
        Spacing (degrees) between observer points in sector. If domain exceeds the
        specified spacing, multiple observer are automatically added across the sector.

    Returns
    -------
    dict
        dictionary holding next satellite overpass estimates for sectors
        (sorted by sector -> satellite -> overpass info)
    """
    from geoips.sector_utils.utils import create_areadefinition_from_yaml

    if not isinstance(satellite_list, list):
        raise TypeError("satellite_list must be type list")
    sector_overpasses = {}
    for yaml_sector in sector_list:
        sector_area_def = create_areadefinition_from_yaml(sectorfile, yaml_sector)
        overpasses = predict_overpass_area_def(
            tlefile,
            sector_area_def,
            satellite_list,
            start_datetime,
            observer_spacing=observer_spacing,
        )
        sector_overpasses[yaml_sector] = overpasses
    return sector_overpasses



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS NetCDF filename production."""

# Python Standard Libraries
import logging

from geoips.filenames.base_paths import PATHS as gpaths
from os.path import join as pathjoin

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "data"
name = "geoips_netcdf_fname"


def call(
    area_def,
    xarray_obj,
    product_names,
    coverage=None,
    output_type="nc",
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    source_dir=None,
    basedir=None,
):
    """Filename formatting for standard GeoIPS-style NetCDF outputs.

    This uses the "assemble_geoips_netcdf_fname" function to appropriately
    assemble the filename from a base directory, product name, source
    name, platform nae, sector name, and product time, to allow reuse
    of this basic filename format from multiple filename formatter
    plugins.
    """
    if basedir is None:
        basedir = gpaths["PRECALCULATED_DATA_PATH"]
    ncdf_fname = assemble_geoips_netcdf_fname(
        basedir=basedir,
        product_name="_".join(product_names),
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        sector_name=xarray_obj.area_definition.area_id,
        product_datetime=xarray_obj.start_datetime,
    )

    return ncdf_fname


def assemble_geoips_netcdf_fname(
    basedir,
    product_name,
    source_name=None,
    platform_name=None,
    sector_name=None,
    product_datetime=None,
    set_subpath=None,
    time_format="%H%M%S",
):
    """Produce full output product path from product / sensor specifications.

    netcdf paths are of the format:
        <basedir>/<product_name>/<source_name>/<platform_name>/
          <sector_name>/date{%Y%m%d}
    netcdf filenames are of the format:
        <date{%Y%m%d>.<time{%H%M%S}>.<platform_name>.<product_name>.
          <sector_name>.nc

    Parameters
    ----------
    basedir : str
        Base directory (additional subdirectories assembled below basedir)
    product_name : str
        Name of product, used in path and filename
    source_name : str
        Name of data source (sensor), used in path and filename
    platform_name : str
        Name of platform (satellite), used in path and filename
    coverage : float
        Image coverage, float between 0.0 and 100.0, used in filename
    product_datetime : datetime
        Datetime object - start time of data used to generate product, used
        in filename
    """
    if set_subpath:
        path = pathjoin(basedir, set_subpath)
    else:
        path = basedir
        if product_name is not None:
            path = pathjoin(path, product_name)
        if source_name is not None:
            path = pathjoin(path, source_name)
        if platform_name is not None:
            path = pathjoin(path, platform_name)
        if sector_name is not None:
            path = pathjoin(path, sector_name)
        if product_datetime is not None:
            path = pathjoin(path, product_datetime.strftime("%Y%m%d"))
    fname = ""
    path_parts = []
    if product_datetime is not None:
        path_parts.extend(
            [
                product_datetime.strftime("%Y%m%d"),
                product_datetime.strftime(time_format),
            ]
        )
    if platform_name is not None:
        path_parts.extend([platform_name])
    if product_name is not None:
        path_parts.extend([product_name])
    if sector_name is not None:
        path_parts.extend([sector_name])
    fname = ".".join(path_parts + ["nc"])

    return pathjoin(path, fname)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Generates all available plugins from all installed GeoIPS packages.

After all plugins have been generated, they are written to a registered_plugins.json
file which contains a dictionary of all the registered GeoIPS plugins across
all plugin repositories.

Call 'python create_plugin_registry.py' to produce registered_plugins.json for
EVERY currently installed plugin package. A separate registered_plugins.json is
created at the top level package directory for each plugin package.
"""

import warnings
import yaml
from importlib import metadata, resources, util, import_module
from inspect import signature
from os.path import (
    basename,
    dirname,
    split,
    splitext,
    exists,
    join as os_join,
    relpath as os_relpath,
)
from os import remove
import re
import logging
import geoips.interfaces
from geoips.errors import PluginRegistryError
import json
from argparse import ArgumentParser


LOG = logging.getLogger(__name__)


def format_docstring(docstring, use_regex=True):
    """Format the provided docstring placement in the plugin registry.

    Found when using the CLI and inspecting the registry, some docstrings are formatted
    in a hard to read manner and look pretty bad. This function will format these
    docstrings to be easily readable, whether obtained via the CLI or manually inspected
    in the plugin registry.

    Parameters
    ----------
    docstring: str
        - The docstring which we are going to format.
    use_regex: bool, optional (default=False)
        - Whether or not we want to apply regex formatting to the docstring. Usually
          recommended as it will replace 'newline' chars but not purposeful
          '.newline' strings.
    """
    if docstring:
        if use_regex:
            # Regex pattern for subbing out "\n" but not ".\n"
            pattern = r"(?<!\.)\n"
            docstring = re.sub(
                pattern,
                " ",
                docstring.strip().replace("\n\n", "\n"),
            )
        else:
            docstring = docstring.strip().replace("\n\n", "\n")
    return docstring


def remove_registries(plugin_packages):
    """Remove all plugin registries if a PluginRegistryError is raised.

    Parameters
    ----------
    plugin_packages: list EntryPoints
        A list of EntryPoints pointing to each installed
        GeoIPS package --> ie.
        [EntryPoint(name='geoips', value='geoips', group='geoips.plugin_packages'), ...]

    Returns
    -------
    None
    """
    # Remove registries is called whenever an improperly formatted plugin or
    # package is encountered.  This is not called until all errors have been
    # collected and reported, to facilitate rapidly identifying and resolving
    # errors.
    LOG.interactive(
        "\n\n\n\nERROR: Removing registries due to improperly formatted plugins.\n"
        "You must fix the error(s) shown below before GeoIPS can operate correctly.\n"
        "Once fixed, please run 'create_plugin_registries' to set up GeoIPS "
        "appropriately\n\n\n"
    )
    # Remove registered_plugins.yaml and registered_plugins.json if they exist
    # for each plugin package.
    for pkg in plugin_packages:
        yaml_plug_path = str(resources.files(pkg.value) / "registered_plugins.yaml")
        json_plug_path = str(resources.files(pkg.value) / "registered_plugins.json")
        if exists(yaml_plug_path):
            remove(yaml_plug_path)
        if exists(json_plug_path):
            remove(json_plug_path)


def registry_sanity_check(plugin_packages, save_type):
    """Check that each plugin package registry has no duplicate lowest depth entries.

    If it does, raise a `PluginRegistryError` for that specific package, then remove all
    plugin registries from each package so the user must fix the error before
    continuing. While this doesn't cause a normal error, duplicate plugins will be
    overwritten by same-named plugin found in the last package-entrypoint.

    Parameters
    ----------
    plugin_packages: list EntryPoints
        A list of EntryPoints pointing to each installed
        GeoIPS package --> ie.
        `[EntryPoint(name='geoips', value='geoips', group='geoips.plugin_packages')]`
    save_type: str
        The file format to save to `[json, yaml]`

    Returns
    -------
    No returns

    Exceptions
    ----------
    PluginRegistryError
        If `error_message` has contents, then raise PluginRegistryError(error_message).
        The `error_message` string will collect and report on all errors within
        this function prior to raising the `PluginRegistryError` to facilitate rapidly
        identifying and resolving errors throughout all plugin packages.
    """
    error_message = ""
    # comp_pkg is the package being compared against. This package is compared
    # against every other available GeoIPS package that is installed.
    for comp_idx, comp_pkg in enumerate(plugin_packages):
        # yaml output is used primarily for testing purposes (since it is more human
        # readable than json), and json output is used for processing. Ensure we
        # can load either option.
        if save_type == "yaml":
            comp_registry = yaml.safe_load(
                open(resources.files(comp_pkg.value) / "registered_plugins.yaml")
            )
        else:
            # json.load is much faster than yaml.safe_load
            comp_registry = json.load(
                open(resources.files(comp_pkg.value) / "registered_plugins.json", "r")
            )
        for pkg_idx, pkg in enumerate(plugin_packages):
            # pkg is the package being compared against comp_pkg. For example, if
            # comp_pkg was 'geoips', then it would compare against recenter_tc,
            # data_fusion, template_basic_plugin, etc.

            # The if statement below checks the index of pkg_idx and comp_idx.
            # If pkg_idk <= comp_idx, that means it's either the same package as
            # comp_pkg, or that the comparison has already been performed.
            if pkg_idx <= comp_idx:
                continue
            # Track sets of plugins by plugin type
            # (schemas, yaml_based, and module_based)
            if save_type == "yaml":
                pkg_registry = yaml.safe_load(
                    open(resources.files(pkg.value) / "registered_plugins.yaml")
                )
            else:
                pkg_registry = json.load(
                    open(resources.files(pkg.value) / "registered_plugins.json", "r")
                )
            for plugin_type in list(pkg_registry.keys()):
                # check the pkg's registry for both yaml-based and module-based plugins
                for interface in comp_registry[plugin_type]:
                    # check each interface of comp_pkg
                    # for each type of plugin (yaml/module)-based
                    if interface in pkg_registry[plugin_type]:
                        # if this interface is also in the pkg_registry, then get the
                        # dictionary of comp_plugins for that interface
                        comp_plugins = comp_registry[plugin_type][interface]
                        for plugin in comp_plugins:
                            # for each plugin in comp_plugins dict
                            if plugin in pkg_registry[plugin_type][interface]:
                                # if this plugin is also in the pkg_registry's
                                # corresponding interface, then retrieve that plugin
                                pkg_plugin = pkg_registry[plugin_type][interface][
                                    plugin
                                ]
                                comp_plugin = comp_registry[plugin_type][interface][
                                    plugin
                                ]
                                if "relpath" in pkg_plugin:
                                    # If 'relpath' is found in the plugin, raise a
                                    # PluginRegistryError, and remove the registries.
                                    # We do this because 'family'  is a top level
                                    # attribute on all plugins, and means you're at
                                    # the lowest depth of the Plugin entry, meaning
                                    # there are two Plugins with Duplicate Names.
                                    error_message += """Error with packages [{}, {}]:
                                        You can't have two Plugins of the same
                                        interface [{}] with the same
                                        plugin name [{}]
                                        pkg relpath: {}
                                        comp relpath: {}""".format(
                                        comp_pkg.value,
                                        pkg.value,
                                        interface,
                                        plugin,
                                        pkg_plugin["relpath"],
                                        comp_plugin["relpath"],
                                    )
                                else:
                                    # If the statement above is false,
                                    # that means the plugin
                                    # we are dealing with is 'Product'-based.
                                    # This means there are subplugins that we
                                    # need to check against their defind
                                    # source names. Grab the comparsion
                                    # Product Plugin.
                                    for sub_plg in comp_plugin:
                                        # Loop through each sub-plugin of the
                                        # comparison product plugin.
                                        if sub_plg in pkg_plugin:
                                            # If this sub-plugin is also in
                                            # the package Product plugin,
                                            # raise a PluginRegistryError
                                            # and remove the registries.
                                            pkg_relpath = "not defined"
                                            if "relpath" in pkg_plugin:
                                                pkg_relpath = pkg_plugin["relpath"]
                                            subplg_relpath = "not defined"
                                            if "relpath" in sub_plg:
                                                subplg_relpath = sub_plg["relpath"]
                                            error_message += """
                                                Error with packages:
                                                [{}, {}]:
                                                You can't have two products of the same
                                                interface [{}] with the same
                                                plugin name [{}] found under
                                                sub_plg name [{}]
                                                relpath: {}
                                                sub_plg relpath: {}\n""".format(
                                                comp_pkg.value,
                                                pkg.value,
                                                interface,
                                                sub_plg,
                                                plugin,
                                                pkg_relpath,
                                                subplg_relpath,
                                            )
    if error_message:
        remove_registries(plugin_packages)
        raise PluginRegistryError(error_message)


def check_plugin_exists(package, plugins, interface_name, plugin_name, relpath):
    """Check if plugin already exists. If it does raise a `PluginRegistryError`.

    Note this only checks for duplicate plugins within a single plugin package.
    The `registry_sanity_check` function is used after all plugins have been
    loaded to identify duplicate plugins across different plugin packages.

    Parameters
    ----------
    package: str
        The GeoIPS package being tested against
    plugins: dict
        A dictionary object of all installed GeoIPS plugins in the current
        plugin package.
    interface_name: str
        A string representing the GeoIPS interface being checked against
    plugin_name: str
        A string representing the name of the plugin within the GeoIPS interface

    Returns
    -------
    error_message : str
        Empty string if no error, appropriate informative error message if
        duplicate plugin found in current plugin package.
    """
    # Check if the passed in plugin_name is already in the current plugin
    # package dictionary for this interface.
    if plugin_name in plugins[interface_name]:
        return f"""\nError in package [{package}]:
            You can not have two Plugins of the same
            interface [{interface_name}] with the same
            name [{plugin_name}] found at
            relpath [{plugins[interface_name][plugin_name]["relpath"]}] and
            relpath [{relpath}]
            """
    return ""


def write_registered_plugins(pkg_dir, plugins, save_type):
    """Write dictionary of all plugins available from installed GeoIPS packages.

    Parameters
    ----------
    pkg_dir: str
        Path in which to write registered_plugins
    plugins: dict
        A dictionary object of all installed GeoIPS package plugins
    save_type: str
        The file format to save to [json, yaml]

    Returns
    -------
    No returns, file written to `pkg_dir`
    """
    if save_type == "yaml":
        reg_plug_abspath = os_join(pkg_dir, "registered_plugins.yaml")
        with open(reg_plug_abspath, "w") as plugin_registry:
            LOG.interactive("Writing %s", reg_plug_abspath)
            yaml.safe_dump(plugins, plugin_registry, default_flow_style=False)
    else:
        reg_plug_abspath = os_join(pkg_dir, "registered_plugins.json")
        with open(reg_plug_abspath, "w") as plugin_registry:
            LOG.interactive("Writing %s", reg_plug_abspath)
            json.dump(plugins, plugin_registry, indent=4)


def create_plugin_registries(plugin_packages, save_type, namespace):
    """Generate all plugin paths associated with every installed GeoIPS packages.

    These paths include schema plugins, module_based plugins
    and normal YAML plugins. After these paths are generated, they are sent
    to parse_plugin_paths, which generates and adds the actual plugins to the
    plugins dictionary.

    Parameters
    ----------
    plugin_packages: list EntryPoints
        A list of EntryPoints pointing to each installed
        GeoIPS package --> ie.
        [EntryPoint(name='geoips', value='geoips', group='geoips.plugin_packages'), ...]
    save_type: str
        The file format to save to [json, yaml]
    namespace: str
        Namespace that your plugin packages fall under. The argument parser defaults
        this value to 'geoips.plugin_packages', but a user can create separate
        namespaces if developing interfaces outside of the main GeoIPS package.
    """
    # It appears when there is *.egg-info directory, it picks that package up
    # twice in the list.  If the same package path exists twice, only keep one
    # of them.  This appears to be a bug with Python 3.9 entry points.
    pkg_dirs = []
    unique_package_entry_points = []
    # Note we only use ep.value (resources.files finds the actual plugins),
    # so we do not need to worry about saving the "wrong" package here.
    # We are actually looping through all the files in each package, so
    # we do not have an entry point for every plugin, just a single entry
    # point for each plugin package.
    for ep in plugin_packages:
        pkg_dir = str(resources.files(ep.value))
        if pkg_dir not in pkg_dirs:
            pkg_dirs += [pkg_dir]
            # Grab the unique package entry points.
            unique_package_entry_points += [ep]

    error_message = ""
    # Loop through only the unique package entry points to avoid duplicate
    # plugin errors.
    for pkg in unique_package_entry_points:
        # This is passed by reference and populated with each call to parse
        # plugin packages.
        plugins = {
            # "schemas": {},
            "text_based": {},
            "yaml_based": {},
            "module_based": {},
        }
        # Track sets of plugins by plugin type
        # (schemas, yaml_based, and module_based)
        package = pkg.value
        LOG.debug("package == " + str(package))
        # We are specifically looping through all files in the ``plugins``
        # directory within the plugin package. Potentially we may want
        # to update this in the future to actually include package_plugins
        # and package_schema entry points, and point the pyproject.toml
        # package_plugins entry point directly to the appropriate directory
        # that holds all plugins, and the package_schema entry point directly
        # to the directory that holds all scheam, so the subdirectories are
        # not hard coded here in the create_plugin_registries code.
        pkg_plugin_path = resources.files(package) / "plugins"
        pkg_dir = str(resources.files(package))
        # Grab all YAML, Python, and txt files within the plugins directory.
        # YAML schema files may be supported in the future.
        yaml_files = pkg_plugin_path.rglob("*.yaml")
        python_files = pkg_plugin_path.rglob("*.py")
        text_files = pkg_plugin_path.rglob("*.txt")
        # Potentially support installing schema into the geoips name space
        # using entry points as well.  Currently unsupported, but this would
        # allow specifying different YAML plugin schema in different
        # repositories.  Currently all supported YAML plugin formats must be
        # specified in the main geoips repo.
        # schema_yaml_path = resources.files(package) / "schema"
        # schema_yamls = schema_yaml_path.rglob("*.yaml")
        # plugin_paths dictionary contains lists of files for each plugin
        # type (ie, yaml based, text based, and module based plugins, and
        # in the future potentially schema)
        plugin_paths = {
            "yaml": sorted(yaml_files),
            "text": text_files,
            # "schemas": schema_yamls,
            "py_files": python_files,
        }
        # `plugins` is passed by reference and populated with all YAML, text, and
        # python plugins found within the current plugin package `package`.
        # This dictionary is formatted appropriately to be written out to the
        # plugin registry file as either a json or YAML output.
        # If any errors are found, append the error message string to the current
        # error_message.  Do not raise an exception until all plugins have been
        # read in, so we can collect and report on all errors at once.
        error_message += parse_plugin_paths(
            plugin_paths, package, pkg_dir, plugins, namespace
        )
        LOG.debug("Available Plugin Types:\n" + str(plugins.keys()))
        LOG.debug(
            "Available YAML Plugin Interfaces:\n" + str(plugins["yaml_based"].keys())
        )
        LOG.debug(
            "Available Module Plugin Interfaces:\n"
            + str(plugins["module_based"].keys())
        )
        # Write the current plugin dictionary to the registered plugins file.
        write_registered_plugins(pkg_dir, plugins, save_type)
    # If error_message is not the empty string, that means we had some errors,
    # so handle appropriately.
    if error_message:
        # Remove all registries to prevent running geoips with an incomplete
        # or corrupt set of plugins.  Force user to resolve errors before
        # proceeding.
        remove_registries(metadata.entry_points(group=namespace))
        # Now raise the error, including the error message with output
        # from every failed plugin/file during the attempted registry process.
        raise PluginRegistryError(error_message)
    # Above error only occurs for duplicates within a single registry.
    # registry_sanity_check will check for duplicates across all
    # registries.
    registry_sanity_check(unique_package_entry_points, save_type)


def parse_plugin_paths(plugin_paths, package, package_dir, plugins, namespace):
    """Parse the plugin_paths provided from the current installed GeoIPS package.

    Then, add them to the plugins dictionary based on the path of the plugin.
    The path contains information as to whether the plugin is a schema, module_based,
    or a normal yaml plugin.

    Parameters
    ----------
    plugin_paths: dict
        A dictionary of filepaths, with keys referring to the type of plugin
    package: str
        The current GeoIPS package being parsed
    package_dir: str
        The path to the current GeoIPS package (for determining relative paths)
    plugins: dict
        A dictionary object of all installed GeoIPS package plugins
    namespace: str
        Namespace that your plugin packages fall under. The argument parser defaults
        this value to 'geoips.plugin_packages', but a user can create separate
        namespaces if developing interfaces outside of the main GeoIPS package.

    Returns
    -------
    error_message: str
        String containing informative error messages from any plugins that
        were improperly formatted.  An exception will be raised at the
        very end if error_message is not the empty string - this allows
        collecting ALL errors throughout the plugin registry process and
        reporting them all at once, to facilitate rapidly identifying and
        resolving errors.
    """
    error_message = ""
    # Loop through each plugin type, ie, text, yaml, module, and later schema.
    for plugin_type in plugin_paths:
        # Loop through each file of the current plugin type.
        for filepath in plugin_paths[plugin_type]:
            filepath = str(filepath)
            # Path relative to the package directory
            relpath = os_relpath(filepath, start=package_dir)
            # plugins is passed by reference, so any new plugins found
            # are added to the plugins dictionary and retained throughout.
            if plugin_type == "yaml":  # yaml based plugins
                error_message += add_yaml_plugin(
                    filepath, relpath, package, plugins["yaml_based"], namespace
                )
            # Ensure we append any errors to the error_message as we go.
            # Exception will not be raised until the very end, when we
            # have collected ALL errors.  This makes it easier to fix
            # all the errors at once.
            elif plugin_type == "text":
                error_message += add_text_plugin(
                    package, relpath, plugins["text_based"]
                )
            # Potentially support schema in the future.
            # elif plugin_type == "schemas":  # schema based yamls
            #     add_schema_plugin(
            #         filepath, abspath, relpath, package, plugins["schemas"]
            #     )
            else:  # module based plugins
                error_message += add_module_plugin(
                    package, relpath, plugins["module_based"]
                )
    # Ensure we return a string error_message with ALL errors appended.
    # This will be raised at the end if error_message has any contents.
    return error_message


def add_yaml_plugin(filepath, relpath, package, plugins, namespace):
    """Add the yaml plugin associated with the filepaths and package to plugins.

    Parameters
    ----------
    filepath: str
        The path of the plugin derived from resouces.files(package) / plugin
    relpath: str
        The relative path to the filepath provided
    package: str
        The current GeoIPS package being parsed
    plugins: dict
        A dictionary object of all installed GeoIPS package plugins
    namespace: str
        Namespace that your plugin packages fall under. The argument parser defaults
        this value to 'geoips.plugin_packages', but a user can create separate
        namespaces if developing interfaces outside of the main GeoIPS package.

    Returns
    -------
    error_message: str
        String containing informative error messages from any plugins that
        were improperly formatted.  An exception will be raised at the
        very end if error_message is not the empty string - this allows
        collecting ALL errors throughout the plugin registry process and
        reporting them all at once, to facilitate rapidly identifying and
        resolving errors.
    """
    for plugin in yaml.load_all(open(filepath, mode="r"), Loader=yaml.SafeLoader):
        plugin["relpath"] = relpath
        plugin["package"] = package

        try:
            interface_name = plugin["interface"]
        except KeyError:
            raise PluginRegistryError(
                f"""No 'interface' level in '{filepath}'.
                    Ensure all required metadata is included."""
            )
        if namespace != "geoips.plugin_packages":
            mod = import_module(package)
            interface_module = getattr(mod.interfaces, f"{interface_name}")
        else:
            interface_module = getattr(geoips.interfaces, f"{interface_name}")

        if interface_name not in plugins.keys():
            plugins[interface_name] = {}

        error_message = ""
        # If the current family is "list", make sure we loop through the list,
        # expanding out each individual product found within the list.
        if plugin["family"] == "list":
            # These are not complete plugins at this stage, only the metadata,
            # so do not validate the plugins here, they will be validated on open.

            # plugin_yaml_to_obj returns the actual plugin object, ie, ProductsPlugin,
            # or SectorsPlugin, etc.
            plg_list = interface_module._plugin_yaml_to_obj(plugin["name"], plugin)

            # plg_list is an e.g. ProductsPlugin object of family list,
            # so within its e.g. plg_list["spec"]["products"] key, we will
            # find a list of all products contained within this single
            # plugin file.
            for yaml_subplg in plg_list["spec"][interface_module.name]:
                # _create_registered_plugin_names will return a list of names
                # found in this single product specification.  Most interfaces
                # this will just return a list of length one, containing
                # [plugin.name], but for products it will return a list of
                # tuples of (source_name, product_name), allowing specifying
                # a list of valid sources within each product spec.
                subplg_names = interface_module._create_registered_plugin_names(
                    yaml_subplg
                )
                # Loop through each of the returned registered plugin names.
                # Give each one its own entry in the plugin registry for easy
                # access.
                for subplg_name in subplg_names:
                    subplg_source = str(subplg_name[0])
                    subplg_product = str(subplg_name[1])
                    if subplg_source not in plugins[interface_name]:
                        plugins[interface_name][subplg_source] = {}
                    # since we are dealing with sub-plugins of a product plugin,
                    # include a couple other pieces of information, such as
                    # product_defaults and source_names.
                    source_names = yaml_subplg["source_names"]
                    pd = None
                    if "product_defaults" in yaml_subplg:
                        pd = yaml_subplg["product_defaults"]
                    family = None
                    if "family" in yaml_subplg:
                        family = yaml_subplg["family"]
                    docstring = None
                    if "docstring" in yaml_subplg:
                        docstring = yaml_subplg["docstring"]
                    # If docstring or family are not specified, and a
                    # product_defaults isn't specified, raise an error
                    # (docstring and family must be defined in a
                    # product_defaults if not defined explicitly).
                    if (not docstring or not family) and not pd:
                        error_message += f"""
                        Error with package '{plugin["package"]}':
                            docstring or family not defined in product,
                            and product_defaults not specified.
                            Must specify docstring and family in either product
                            or product_defaults.
                            interface '{interface_module.name}'
                            plugin name '{plugin["name"]}'
                            pkg relpath: '{plugin["relpath"]}'\n"""
                        continue

                    # I think filling in the nulls should be handled at the CLI
                    # and/or interface level, so leave family/docstring as None
                    # in the registry.
                    # There is no way to guarantee the product_defaults are
                    # available in the same plugin package as the current plugin,
                    # so pulling the family and docstring from the product defaults
                    # when creating the registry will not always work.
                    # If we are concerned about efficiency with opening product and
                    # product_defaults with every plugin access, we could do a
                    # second pass in the plugin registry creation to fill in the
                    # nulls (but we will not worry about that yet)
                    # if not docstring or not family:
                    #     # if the yaml_sub_plg doesn't include a docstring, grab its
                    #     # product_defaults docstring
                    #     if (
                    #         "product_defaults" not in plugins
                    #         or pd not in plugins["product_defaults"]
                    #     ):
                    #         LOG.error(
                    #             f"""Product defaults '{pd}' does not exist.
                    #                   Using 'undefined' docstring.
                    #                   Need to figure out how to pull product defaults
                    #                   from a different plugin package at some point."""  # NOQA
                    #         )
                    #     else:
                    #         if not docstring:
                    #             docstring = plugins["product_defaults"][pd]["docstring"]  # NOQA
                    #         if not family:
                    #             family = plugins["product_defaults"][pd]["family"]
                    plugins[interface_name][subplg_source][subplg_product] = {
                        "docstring": format_docstring(docstring),
                        "family": family,
                        "interface": interface_module.name,
                        "package": plugin["package"],
                        "plugin_type": "yaml_based",
                        "product_defaults": pd,
                        "source_names": source_names,
                        "relpath": plugin["relpath"],
                    }
        else:
            error_message += check_plugin_exists(
                package, plugins, interface_name, plugin["name"], relpath
            )

            # If this is not of family list, just set a single entry for
            # current plugin name.
            # Since this is not a product plugin, we can ensure that these top-level
            # attributes should exist. Don't include product_defaults or source_names in
            # this info, because it doesn't apply to this type of plugin.
            plugins[interface_name][plugin["name"]] = {
                "docstring": format_docstring(plugin["docstring"]),
                "family": plugin["family"],
                "interface": plugin["interface"],
                "package": package,
                "plugin_type": "yaml_based",
                "relpath": relpath,
            }
    return error_message


def add_text_plugin(package, relpath, plugins):
    """Add all txt plugins into plugin registries.

    Parameters
    ----------
    package: str
        The current GeoIPS package being parsed
    relpath: str
        The relpath path to the module plugin
    plugins: dict
        A dictionary object of all installed GeoIPS package plugins

    Returns
    -------
    error_message: str
        String containing informative error messages from any plugins that
        were improperly formatted.  An exception will be raised at the
        very end if error_message is not the empty string - this allows
        collecting ALL errors throughout the plugin registry process and
        reporting them all at once, to facilitate rapidly identifying and
        resolving errors.
    """
    # Eventually, we will add interface, family, and name to the text files
    # themselves, in which case we will pull the appropriate information out
    # of the attributes included in the comments at the beginning of the text
    # file. Note I added these comments to the current ascii_palettes, though
    # they are not yet used.

    # For now, use the basename of the filename as the "name"
    text_name = splitext(basename(relpath))[0]

    # For now, use the last directory name as the interface name.
    interface_name = split(dirname(relpath))[-1]
    error_message = ""
    if interface_name not in plugins:
        plugins[interface_name] = {}
    plugins[interface_name][text_name] = {"package": package, "relpath": relpath}
    # For now we have no error messages for text plugins, it will always be
    # an empty string.  But return it anyway.
    return error_message


# def add_schema_plugin(filepath, abspath, relpath, package, plugins):
#     """Add the schema plugin associated with the filepaths and package to plugins.
#
#     Parameters
#     ----------
#     filepath: str
#         The path of the plugin derived from resouces.files(package) / schema
#     abspath: str
#         The absolute path to the filepath provided
#     relpath: str
#         The relative path to the filepath provided
#     package: str
#         The current GeoIPS package being parsed
#     plugins: dict
#         A dictionary object of all installed GeoIPS package plugins
#     """
#     import numpy as np
#
#     split_path = np.array(filepath.split("/"))
#     interface_idx = np.argmax(split_path == "schema") + 1
#     interface_name = split_path[interface_idx]
#     if interface_name not in plugins.keys():
#         plugins[interface_name] = {}
#     plugin = yaml.safe_load(open(filepath, mode="r"))
#     plugin["abspath"] = abspath
#     plugin["relpath"] = relpath
#     plugin["package"] = package
#     plugins[interface_name][plugin["$id"]] = plugin
#     # plugins[interface_name].append(
#     #     {plugin["$id"]: {"$id": plugin["$id"], "abspath": abspath}}
#     # )


def add_module_plugin(package, relpath, plugins):
    """Add the module plugin associated with the filepaths and package to plugins.

    Parameters
    ----------
    package: str
        The current GeoIPS package being parsed
    relpath: str
        The relpath path to the module plugin
    plugins: dict
        A dictionary object of all installed GeoIPS package plugins

    Returns
    -------
    error_message: str
        String containing informative error messages from any plugins that
        were improperly formatted.  An exception will be raised at the
        very end if error_message is not the empty string - this allows
        collecting ALL errors throughout the plugin registry process and
        reporting them all at once, to facilitate rapidly identifying and
        resolving errors.
    """
    error_message = ""
    if "__init__.py" in relpath:
        return error_message
    module_name = splitext(basename(relpath))[0]
    # We need the full path to the module in order
    # for relative imports to work within modules.
    module_path = splitext(relpath.replace("/", "."))[0]
    module_path = f"{package}.{module_path}"
    abspath = resources.files(package) / relpath

    spec = util.spec_from_file_location(module_path, abspath)

    module = util.module_from_spec(spec)
    # Attempting importing module, catch ImportError
    # We have to fix these to be able to import the module to
    # see if 'interface' is defined, in order to see if it
    # is a properly formatted python module..
    try:
        spec.loader.exec_module(module)
    except ImportError as resp:
        LOG.exception(resp)
        error_message += f"""\nError {str(resp)}:
                         Failed importing '{module_name}' in
                         package '{package}'
                         at relpath '{relpath}'\n"""
        return error_message
    # Try to get "interface" variable from the module.  This is required
    # on ALL files within the python module based plugins directory, to
    # ensure create_plugin_registries can explicitly tell whether a file
    # is properly formatted or not.  Files that are not full plugins must
    # be specified with "interface = None" (identifying as a python module
    # that should NOT be included in the python registry), and full GeoIPS
    # plugins must include interface, family, and name variables at the top
    # level.
    try:
        interface_name = module.interface
    except AttributeError:
        error_message += f"""\nError,
            'interface' top level variable missing in
            module '{module_name}' in
            package '{package}'
            at relpath '{relpath}'

            * must specify 'interface' variable at the
              top level of ALL python modules within the
              plugins subdirectory.

              * FOR VALID GEOIPS PLUGINS:
                'interface', 'family', and 'name' must all be specified
                as variables at the top level.

              * FOR HELPER MODULES WITHIN THE plugins SUBDIRECTORY
                'interface = None' must be specified at the top level for modules
                within the plugins subdirectory that are not intended to be
                GeoIPS plugins on their own."""
        return error_message
    # If interface is None, then legitimately skip the module.
    # We want to skip this first, before we test anything else.
    # If it is not a plugin, we don't care if there are other
    # errors in it at this stage (ie, avoid unnecessary unrelated
    # catastrophic failures)
    if not interface_name:
        LOG.interactive(
            f"Skipping module '{module_name}' from '{package}', "
            "interface_name is 'None'"
        )
        return error_message
    # If we get here, it should be a full GeoIPS plugin, so it must include both
    # name and family variables/attributes.
    try:
        name = module.name
        family = module.family
    except AttributeError:
        error_message += f"""\nError, 'family' or 'name' top level variable missing in
            module '{module_name}' in package '{package}'
            at relpath '{relpath}'
            must specify 'interface', 'family', and 'name' variables at the
            top level of ALL module based plugins."""
        return error_message
    # If the current interface_name is not in the plugins dictionary yet, add it
    # as an empty dictionary.
    if interface_name not in plugins.keys():
        plugins[interface_name] = {}
    # Check_plugin_exists will return a text error message if there are any errors
    # rather than raising an exception.  This allows collecting all errors as
    # we go, and reporting once at the end with an error message including ALL
    # errors found across all plugins in all plugin packages.  Append the new
    # error message to the error messages that have already been collected.
    error_message += check_plugin_exists(
        package, plugins, interface_name, name, relpath
    )
    # Add info shown below obtained from the module plugin. Every module plugin
    # is required to have these entries in the registry to be considered a valid
    # plugin.
    plugins[interface_name][name] = {
        "docstring": format_docstring(module.__doc__),
        "family": family,
        "interface": interface_name,
        "package": package,
        "plugin_type": "module_based",
        "signature": str(signature(module.call)),
        "relpath": relpath,
    }
    if interface_name == "readers":
        if hasattr(module, "source_names"):
            plugins[interface_name][name]["source_names"] = module.source_names
        else:
            warnings.warn(
                (
                    f"Plugin package '{package}'s reader"
                    f" plugin '{name}' is using a deprecated source_names "
                    "implementation. Please add a module-level 'source_names' "
                    "attribute to this plugin and re-run "
                    "'create_plugin_registries'. This will be fully deprecated "
                    "when GeoIPS v2.0.0 is released."
                ),
                DeprecationWarning,
                stacklevel=2,
            )
            plugins[interface_name][name]["source_names"] = ["Unspecified"]
    del module
    # Return the final error message - an exception will be raised at the very
    # end after collecting and reporting on all errors if there were any errors
    # during plugin registry creation.
    return error_message


def get_parser():
    """Create the ArgumentParser for main."""
    description = (
        "Creates Plugin Registries for all installed GeoIPS packages. "
        "The registries will be written to the root directory of each installed "
        "package. The registries will be named either 'registered_plugins.json' "
        "or 'registered_plugins.yaml' depending on which format is chosen. "
        "For additional information on GeoIPS plugin registries please refer to "
        "the GeoIPS documentation."
    )
    parser = ArgumentParser(
        prog="create_plugin_registries",
        description=description,
    )
    parser.add_argument(
        "-s",
        "--save_type",
        type=str.lower,
        default="json",
        choices=["json", "yaml"],
        help="Format to write registries to. This will also be the file extension.",
    )
    parser.add_argument(
        "-p",
        "--package_name",
        type=str.lower,
        default=None,
        help="Package name to create registries for. If not specified, run on all.",
    )
    parser.add_argument(
        "-n",
        "--namespace",
        type=str,
        default="geoips.plugin_packages",
        help=(
            "Namespace that your plugin packages fall under. Defaults to geoips, but a "
            "user can create separate namespaces if developing interfaces outside of "
            "the main GeoIPS package."
        ),
    )
    return parser



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test script for representative product comparisons."""

import logging
from PIL import Image
import numpy as np
from os import makedirs
from os.path import exists, join, splitext

from geoips.geoips_utils import get_numpy_seeded_random_generator

from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "image"


def get_test_files(test_data_dir):
    """Return a series of compare vs output image paths for testing purposes."""
    savedir = join(test_data_dir, "scratch", "unit_tests", "test_images/")
    if not exists(savedir):
        makedirs(savedir)

    thresholds = ["lenient", "medium", "strict"]
    # thresholds is used for naming files.
    # Relates to thresholds [0.1, 0.05, 0.0]
    compare_files = []
    test_files = []

    predictable_random = get_numpy_seeded_random_generator()

    for threshold in thresholds:
        for i in range(3):
            comp_arr = predictable_random.rand(100, 100, 3)
            test_arr = np.copy(comp_arr)
            if i == 1:
                rand = predictable_random.randint(0, 100)
                test_arr[rand][:] = predictable_random.rand(3)
            elif i == 2:
                test_arr = predictable_random.rand(100, 100, 3)
            comp_img = Image.fromarray((comp_arr * 255).astype(np.uint8))
            test_img = Image.fromarray((test_arr * 255).astype(np.uint8))
            comp_file = join(savedir, f"comp_img_{threshold}{str(i)}.png")
            test_file = join(savedir, f"test_img_{threshold}{str(i)}.png")
            comp_img.save(comp_file)
            test_img.save(test_file)
            compare_files.append(comp_file)
            test_files.append(test_file)
    return compare_files, test_files


def perform_test_comparisons(plugin, compare_files, test_files):
    """Test the comparison of two images with the Image Output Checker."""
    threshold_floats = [0.1, 0.05, 0.0]
    for threshold in threshold_floats:
        for path_idx in range(len(compare_files)):
            retval = plugin.module.outputs_match(
                plugin,
                test_files[path_idx],
                compare_files[path_idx],
                threshold,
            )
            if path_idx % 3 == 0:
                assert retval is True
            else:
                assert retval is False


def correct_file_format(fname):
    """Determine if fname is an image file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is an image file, False otherwise.
    """
    if splitext(fname)[-1] in [".png", ".jpg", ".jpeg"]:
        return True
    return False


def outputs_match(
    plugin,
    output_product,
    compare_product,
    threshold=gpaths["OUTPUT_CHECKER_THRESHOLD_IMAGE"],
):
    """Use PIL and numpy to compare two images.

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding geotiff OutputCheckerPlugin that has access to needed methods
    output_product : str
        Current output product
    compare_product : str
        Path to comparison product
    threshold: float, default=0.05
        Threshold for the image comparison. Argument to pixelmatch.
        Between 0 and 1, with 0 the most strict comparison, and 1 the most lenient.

    Returns
    -------
    bool
        Return True if images match, False if they differ
    """
    exact_out_diffimg_fname = plugin.get_out_diff_fname(
        compare_product, output_product, flag="exact_"
    )
    out_diffimg_fname = plugin.get_out_diff_fname(compare_product, output_product)

    try:
        from pixelmatch.contrib.PIL import pixelmatch
    except ImportError as e:
        raise ImportError(
            "Comparing images using GeoIPS requires a python library called pixelmatch"
            + " (link: https://pypi.org/project/pixelmatch/)."
            + "You can install it directly or install the test extra"
            + " eg. pip install geoips[test]"
        ) from e

    from geoips.commandline.log_setup import log_with_emphasis

    LOG.info("**Comparing output_product vs. compare product")
    # Open existing images.
    out_img = Image.open(output_product)
    comp_img = Image.open(compare_product)
    diff_img = Image.new(mode="RGB", size=comp_img.size)
    exact_diff_img = Image.new(mode="RGB", size=comp_img.size)
    # Compute the pixel diff between the two images.
    if np.array(comp_img).shape == np.array(out_img).shape:
        diff_arr = np.abs(np.array(comp_img) - np.array(out_img))
    # If shapes of arrays do not match, pixel diff can not be performed.
    # Print the names of the two images and associated shapes, and return False.
    else:
        log_with_emphasis(
            LOG.interactive,
            "BAD Images NOT match exactly, different sizes",
        )
        message = f"output_product: {np.array(out_img).shape} {output_product}"
        log_with_emphasis(LOG.interactive, message)
        message = f"compare_product: {np.array(comp_img).shape} {compare_product}"
        log_with_emphasis(LOG.interactive, message)
        return False
    # Don't bother doing pixelmatch if the arrays are identical.
    # Make sure the min/max values are both 0 in the diff image.
    # If there happen to be any nans in the diff_arr, this test will fail, because
    # diff_arr.min/max return nan if there are any nans in the array.
    # For the record, I don't think PIL Image would have nan values, but
    # this should be safe (since NaNs will result in moving onto the pixelmatch tests).
    if diff_arr.min() == 0 and diff_arr.max() == 0:
        # If the images match exactly, just output to GOOD comparison log to info level
        # (only bad comparisons to interactive level)
        log_with_emphasis(
            LOG.info,
            "GOOD Images match exactly",
        )
        return True

    # Determine the number of pixels that are mismatched
    LOG.info("Using pixelmatch threshold %s", threshold)
    thresholded_retval = pixelmatch(
        out_img, comp_img, diff_img, includeAA=True, alpha=0.33, threshold=threshold
    )
    exact_retval = pixelmatch(
        out_img, comp_img, exact_diff_img, includeAA=True, alpha=0.33, threshold=0
    )
    # Write out the exact image difference.
    LOG.info("**Saving exact difference image")
    exact_diff_img.save(exact_out_diffimg_fname)
    # Write out the thresholded image difference.
    LOG.info("**Saving thresholded difference image")
    diff_img.save(out_diffimg_fname)
    LOG.info("**Done running compare")

    bad_inds = np.where(diff_arr != 0)
    num_mismatched_exactly = len(diff_arr[bad_inds])
    num_mismatched_by_threshold = thresholded_retval
    num_total_pixels = diff_img.size[0] * diff_img.size[1]
    bad_threshold_pct = num_mismatched_by_threshold / num_total_pixels
    bad_exact_pct = num_mismatched_exactly / num_total_pixels

    # If the images do not match exactly, print the output image, comparison image,
    # and exact diff image to log, for easy viewing.  Return False.

    # NOTE: We should discuss what values we'd like to set, or be dynamic for testing
    # bad_threshold_pct and bad_exact_pct. This is just a first value and will likely
    # change once we've decided what percentage of failures we'll allow.
    if bad_threshold_pct > (threshold / 1000):
        log_with_emphasis(
            LOG.interactive,
            "BAD Images do NOT match within tolerance",
            f"np.where(diff_arr != 0): {bad_inds}",
            f"diff_arr[bad_inds]: {diff_arr[bad_inds]}",
            f"{thresholded_retval} mismatched pixels "
            f"exceeding pixelmatch threshold {threshold}",
            f"{exact_retval} mismatched pixels exceeding pixelmatch threshold 0"
            f"{len(diff_arr[bad_inds])} mismatched exact of {num_total_pixels} pixels",
            f"percentage diff exact {bad_exact_pct}",
            f"percentage diff thresholded {bad_threshold_pct}",
            f"output_product: {output_product}",
            f"compare_product: {compare_product}",
            f"exact diff image: {exact_out_diffimg_fname}",
            f"thresholded diff image: {out_diffimg_fname}",
        )
        return False
    else:
        log_with_emphasis(
            LOG.interactive,
            "GOOD Images match within tolerance",
            f"{len(diff_arr[bad_inds])} mismatched exact of {num_total_pixels} pixels",
            f"{thresholded_retval} mismatched pixels "
            f"exceeding pixelmatch threshold {threshold}",
        )
        log_with_emphasis(
            LOG.info,
            "GOOD Images match within tolerance",
            f"np.where(diff_arr != 0): {bad_inds}",
            f"diff_arr[bad_inds]: {diff_arr[bad_inds]}",
            f"{thresholded_retval} mismatched pixels "
            f"exceeding pixelmatch threshold {threshold}",
            f"{exact_retval} mismatched pixels exceeding pixelmatch threshold 0",
            f"{len(diff_arr[bad_inds])} mismatched exact of {num_total_pixels} pixels",
            f"percentage diff exact {bad_exact_pct}",
            f"percentage diff thresholded {bad_threshold_pct}",
            f"output_product: {output_product}",
            f"compare_product: {compare_product}",
            f"exact diff image: {exact_out_diffimg_fname}",
            f"thresholded diff image: {out_diffimg_fname}",
        )

    return True


def call(
    plugin,
    compare_path,
    output_products,
    threshold=gpaths["OUTPUT_CHECKER_THRESHOLD_IMAGE"],
):
    """Compare the "correct" imagery found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding image OutputCheckerPlugin that has access to needed methods
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path
    threshold: float, default=0.05
        Threshold for the image comparison. Argument to pixelmatch.
        Between 0 and 1, with 0 the most strict comparison, and 1 the most lenient.

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    retval = plugin.compare_outputs(
        compare_path,
        output_products,
        threshold=threshold,
    )
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing Infrared algorithm colormap."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "dmw_wind_levels"


def call(data_range=[0, 3], pressure_range_legend=None):
    """Colormap for displaying multi-level derived motion winds.

    Each level is assigned an integer, with 0 being unassigned.

    Parameters
    ----------
    data_range : list, optional
        Min and max value for colormap, by default [0, 3]
    pressure_range_legend : list, optional
        List of strings that are used for setting the cbar tick labels, by default None

    Returns
    -------
    dict
        Dictionary of matplotlib plotting parameters, to ensure consistent image output
    """
    if not pressure_range_legend:
        pressure_range_legend = ["Unassigned", "Low", "Mid", "High"]
    min_lev = int(data_range[0])
    max_lev = int(data_range[1])

    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [(-0.5, 0.5), (0.5, 1.5), (1.5, 2.5), (2.5, 3.5)]
    transition_colors = [
        ("grey", "grey"),
        ("yellow", "yellow"),
        ("cyan", "cyan"),
        ("magenta", "magenta"),
    ]

    # special selection of label

    ticks = [0, 1, 2, 3]

    # selection of min and max values for colormap if needed
    min_lev = transition_vals[0][0]
    max_lev = transition_vals[-1][1]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "DMW_cmap", min_lev, max_lev, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_lev, vmax=max_lev)

    cbar_label = None

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = pressure_range_legend
    mpl_boundaries = None

    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_lev, max_lev
    return mpl_colors_info



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `describe plugin` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from geoips import interfaces
from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest
from geoips.geoips_utils import get_numpy_seeded_random_generator


class TestGeoipsDescribePlugin(BaseCliTest):
    """Unit Testing Class for Describe Plugin Command."""

    rand_threshold = 0.10
    random_generator = get_numpy_seeded_random_generator()

    @property
    def command_combinations(self):
        """A stochastic list of commands used for the GeoipsDescribePlugin command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "describe"]
            # validate all plugins from package all packages
            for interface_name in interfaces.__all__:
                interface = getattr(interfaces, interface_name)
                # If there happen to be no plugins for a given interface, just skip, do
                # not fail catastrophically. This allows defining interfaces in the
                # geoips repo, even if there are no plugins of that interface directly
                # in the main geoips repo.
                if (
                    interface_name
                    not in interface.plugin_registry.registered_plugins[
                        interface.interface_type
                    ]
                ):
                    continue
                interface_registry = interface.plugin_registry.registered_plugins[
                    interface.interface_type
                ][interface_name]
                for idx, plugin_name in enumerate(interface_registry):
                    interface_name = interface_name.replace("_", "-")
                    for alias in self.alias_mapping[interface_name] + [interface_name]:
                        # Randomly select some plugins to perform 'describe plugin' on.
                        # Doing this because there are too many plugins to test in a
                        # timely manner
                        if interface.name == "products":
                            for subplg_name in interface_registry[plugin_name]:
                                do_describe_plugin = (
                                    self.random_generator.random() < self.rand_threshold
                                )
                                if do_describe_plugin or idx == 0:
                                    combined_name = f"{plugin_name}.{subplg_name}"
                                    self._cmd_list.append(
                                        base_args + [alias, combined_name],
                                    )
                        else:
                            do_describe_plugin = (
                                self.random_generator.random() < self.rand_threshold
                            )
                            if do_describe_plugin or idx == 0:
                                self._cmd_list.append(
                                    base_args + [alias, plugin_name],
                                )
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non_existent_interface
            self._cmd_list.append(base_args + ["non_existent_interface", "Infrared"])
            # Add argument list with non_existent_plugin
            self._cmd_list.append(base_args + ["algorithms", "non_existent_plugin"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips describe plugin ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # An error occurred using args. Assert that args is not valid and check the
        # output of the error.
        err_str = "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
        assert err_str in error

    def check_output(self, args, output):
        """Ensure that the 'geoips describe plugin ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            usg_str = (
                "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
            )
            assert usg_str in output
        else:
            # Checking that output from geoips describe plugin command is valid
            interface_name = args[3]
            expected_outputs = [
                "Docstring",
                "Family",
                "Interface",
                "Package",
                "Plugin Type",
                "Relative Path",
            ]
            if interface_name == "products":
                expected_outputs.append("source_names")
            elif interface_name in interfaces.module_based_interfaces:
                expected_outputs.append("signature")
            for output_item in expected_outputs:
                assert f"{output_item}:" in output


test_sub_cmd = TestGeoipsDescribePlugin()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips describe plugin ...' commands.

    This test covers every valid combination of commands for the 'geoips describe
    plugin' command. We also test invalid commands, to ensure that the proper help
    documentation is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'describe', 'plugin'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args=args)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test Pydantic base models for Order-Based Procflow."""

# Python Standard Libraries
import copy
import json

# Third-Party Libraries
import pytest
from pydantic import Field, ValidationError

# GeoIPS Libraries
from geoips.pydantic import bases


@pytest.mark.parametrize(
    "valid_identifier",
    [
        "variable",
        "_underscore_prefixed_variable",
        "variable_alpha_numeric123",
        "_",
        "valid_variable_with_underscore",
    ],
    ids=[
        "simple identifier",
        "underscore-prefixed",
        "alphanumeric",
        "single underscore",
        "contains underscore",
    ],
)
def test_good_valid_python_identifier(valid_identifier):
    """Tests python_identifier call against multiple valid Python identifiers."""
    assert bases.python_identifier(valid_identifier) == valid_identifier


@pytest.mark.parametrize(
    "invalid_identifier, expected_error",
    [
        ("123_starts_with_number", "is not a valid Python identifier"),
        ("invalid_$_symbol", "is not a valid Python identifier"),
        ("variable with_space", "is not a valid Python identifier"),
        ("", "is not a valid Python identifier"),
        ("class", "is a reserved Python keyword"),
    ],
    ids=[
        "starts with number",
        "contains invalid special character",
        "contains space",
        "empty variable name",
        "uses reserved keyword",
    ],
)
def test_bad_invalid_python_identifier(invalid_identifier, expected_error):
    """Tests python_identifier call against multiple invalid Python identifiers."""
    with pytest.raises(ValueError) as exec_info:
        bases.python_identifier(invalid_identifier)

    assert isinstance(exec_info.value, ValueError), "Exception raised is not ValueError"
    error_message = str(exec_info.value)
    assert (
        expected_error in error_message
    ), f"{error_message} does not match {expected_error}"


# Test PrettyBaseModel
class MockPrettyBaseModel(bases.PrettyBaseModel):
    """Test PrettyBaseModel to test __str__method of PrettyBaseModel."""

    plugin_type: str = Field(description="name of the plugin type")
    plugin_name: str = Field(description="name of the plugin")


def test_good_pretty_base_model_str():
    """Test if the PrettyBaseModel returns JSON data with two-space indentation."""
    test_model = MockPrettyBaseModel(plugin_type="Reader", plugin_name="abi_netcdf")

    string_representation_of_model = str(test_model)
    expected_json_format = json.dumps(
        {"plugin_type": "Reader", "plugin_name": "abi_netcdf"}, indent=2
    )

    assert string_representation_of_model == expected_json_format


def test_bad_pretty_base_model_missing_required_field():
    """Test PrettyBaseModel for missing required fields."""
    with pytest.raises(ValidationError) as exec_info:
        MockPrettyBaseModel(plugin_type="Reader")

    error_info = exec_info.value.errors()
    assert len(error_info) == 1
    assert error_info[0]["loc"] == ("plugin_name",)
    assert error_info[0]["type"] == "missing"


def test_bad_pretty_base_model_invalid_field_type():
    """Test PrettyBaseModel for missing required fields."""
    with pytest.raises(ValidationError) as exec_info:
        MockPrettyBaseModel(plugin_type="Reader", plugin_name=123)

    error_info = exec_info.value.errors()
    assert len(error_info) == 1
    assert error_info[0]["loc"] == ("plugin_name",)
    assert error_info[0]["type"] == "string_type"


# Test PluginModel


def test_good_plugin_valid_instance(valid_plugin_data):
    """Test PluginModel with valid inputs."""
    plugin = bases.PluginModel(**valid_plugin_data)

    assert plugin.interface == valid_plugin_data["interface"]
    assert plugin.family == valid_plugin_data["family"]
    assert plugin.name == valid_plugin_data["name"]
    assert plugin.docstring == valid_plugin_data["docstring"]
    assert plugin.description == valid_plugin_data["description"]
    assert plugin.package == valid_plugin_data["package"]
    assert plugin.relpath == valid_plugin_data["relpath"]
    assert plugin.abspath == valid_plugin_data["abspath"]


def test_bad_plugin_invalid_instance_additional_field(valid_plugin_data):
    """Test PluginModel with additional field."""
    invalid_data = valid_plugin_data.copy()
    # adding an extra field
    invalid_data["unexpected_field"] = "unexpected_value"

    with pytest.raises(ValidationError) as exec_info:
        bases.PluginModel(**invalid_data)

    error_info = exec_info.value.errors()
    assert any(
        err["type"] == "extra_forbidden" for err in error_info
    ), "expected 'extra_forbidden' error type"
    assert "unexpected_field" in str(
        exec_info.value
    ), "Unexpected field should be mentioned in the error"


def test_good_plugin_model_valid_interfaces(valid_plugin_data, valid_interfaces):
    """Test PluginModel's valid_interface() method with valid interfaces."""
    model = bases.PluginModel(**valid_plugin_data)
    assert model.interface in valid_interfaces


def test_bad_plugin_model_valid_interfaces(valid_plugin_data, valid_interfaces):
    """Test PluginModel's  valid_interface() method with invalid interfaces."""
    invalid_interface = "invalid_interface_name"

    assert (
        invalid_interface not in valid_interfaces
    ), f"'{invalid_interface}' should not be a valid interface."

    invalid_data = valid_plugin_data.copy()
    invalid_data["interface"] = invalid_interface

    with pytest.raises(ValidationError, match="Invalid interface"):
        bases.PluginModel(**invalid_data)


# Parameterized test input for a valid description test
@pytest.mark.parametrize(
    "docstring",
    ["This is a valid docstring."],
)
def test_good_plugin_model_docstring(valid_plugin_data, docstring):
    """Test PluginModel with a good docstring."""
    data = copy.deepcopy(valid_plugin_data)
    data["docstring"] = docstring
    model = bases.PluginModel(**data)
    assert model.docstring == docstring


# Parameterized test input for multiple invalid descriptions
@pytest.mark.parametrize(
    "invalid_docstring",
    [
        (
            "This is a \n a multiline docstring.",
            "The description must be one line only.",
        ),
        ("Docstring with missing period", "The docstring must end with a period"),
        (
            "docstring starting with lower case and ending with no period",
            "The Docstring should start with a capital letter and end with a period",
        ),
    ],
    ids=[
        "Multiline docstring",
        "Missing period docstring",
        "Start with Capital & end in period",
    ],
)
def test_bad_plugin_model_description(valid_plugin_data, invalid_docstring):
    """Test PluginModel with invalid docstring use cases."""
    data = copy.deepcopy(valid_plugin_data)
    data["description"] = invalid_docstring

    with pytest.raises(ValueError) as exec_info:
        bases.PluginModel(**data)

    error_info = exec_info.value.errors()
    assert any(error["loc"] == ("description",) for error in error_info)
    assert any("valid string" in error["msg"] for error in error_info)


@pytest.mark.parametrize(
    "docstring_input, expected_description",
    [
        (
            {"docstring": "First line. \n Second line.", "description": None},
            "First line.",
        ),
    ],
    ids=[
        "Description is None",
    ],
)
def test_bad_plugin_model_set_description(
    valid_plugin_data, docstring_input, expected_description
):
    """Test PluginModel's set_description() method when description is set to None."""
    data = copy.deepcopy(valid_plugin_data)
    data.update(docstring_input)
    model = bases.PluginModel(**data)

    assert model.description == expected_description, (
        f"Expected description: {expected_description},"
        f"but got this: {model.docstring}"
    )


@pytest.mark.parametrize(
    "docstring_input, expected_description",
    [
        (
            {
                "docstring": "First line. \n Second line.",
            },
            "First line.",
        ),
    ],
    ids=[
        "Description not provided",
    ],
)
def test_bad_plugin_model_set_description_without_input_field(
    valid_plugin_data, docstring_input, expected_description
):
    """Test PluginModel's set_description() method when description field is missing."""
    data = copy.deepcopy(valid_plugin_data)
    data.pop("description", None)
    data.update(docstring_input)
    model = bases.PluginModel(**data)
    assert model.description == expected_description, (
        f"Expected description: {expected_description},"
        f"but got this: {model.docstring}"
    )


def test_bad_plugin_model_validate_one_line_description_length(valid_plugin_data):
    """Test PluginModel's description validator against more than 72 characters."""
    data = copy.deepcopy(valid_plugin_data)
    data["description"] = (
        "This description is intentionally crafted to exceed seventy-two characters."
    )

    with pytest.warns(
        FutureWarning, match="Description cannot be more than 72 characters."
    ):
        model = bases.PluginModel(**data)

    assert model.description


def test_bad_plugin_model_validate_one_line_description_multi_line(valid_plugin_data):
    """Test PluginModel's description validator against multi line input."""
    data = copy.deepcopy(valid_plugin_data)
    data["description"] = "This description is a \n multi line one."

    with pytest.warns(FutureWarning, match="Description must be a single line."):
        model = bases.PluginModel(**data)

    assert model.description



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module for generating specific colormaps on the fly."""
# Installed Libraries
import logging
import ast

LOG = logging.getLogger(__name__)


def set_matplotlib_colors_rgb():
    """
    Create matplotlib Colors parameters dictionary.

    For rgb imagery, we require no color information (it is entirely
    specified by the RGB(A) arrays).

    Returns
    -------
    mpl_colors_info : dict
        Specifies matplotlib Colors parameters for use in both plotting
        and colorbar generation. For RGBA arrays, all fields are "None".
    """
    mpl_colors_info = {
        "cmap": None,
        "norm": None,
        "cbar_ticks": None,
        "cbar_tick_labels": None,
        "cbar_label": None,
        "boundaries": None,
        "cbar_spacing": "proportional",
        "colorbar": False,
    }
    return mpl_colors_info


def set_matplotlib_colors_standard(
    data_range, cmap_name="Greys", cbar_label=None, create_colorbar=True
):
    """
    Set the matplotlib colors information.

    For use in colorbar and image production.

    Parameters
    ----------
    data_range : list
        the minimum and maximum value for the data range
        [min_val, max_val]
    cmap_name : str, default='Greys'
        Specify the standard matplotlib colormap
    cbar_label : str, optional
        If specified, use cbar_label string as colorbar label
    create_colorbar : bool, default=True
        Specify whether the image should contain a colorbar or not

    Returns
    -------
    mpl_colors_info : dict
        Specifies matplotlib Colors parameters for use in both plotting
        and colorbar generation.
        See geoips.image_utils.mpl_utils.create_colorbar for field
        descriptions.
    """
    min_val = data_range[0]
    max_val = data_range[1]
    from matplotlib import pyplot as plt

    # cmap = cm.ScalarMappable(norm=colors.NoNorm(), cm.get_cmap(cmap_name))
    mpl_cmap = plt.get_cmap(cmap_name)

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_val, vmax=max_val)
    mpl_ticks = [min_val, max_val]

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": mpl_ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": create_colorbar,
    }

    return mpl_colors_info


def set_mpl_colors_info_dict(
    cmap,
    norm,
    cbar_ticks,
    cbar_tick_labels=None,
    boundaries=None,
    cbar_label=None,
    cbar_spacing="proportional",
    create_colorbar=True,
    cbar_full_width=False,
):
    """
    Create the mpl_colors_info dictionary directly from passed arguments.

    Parameters
    ----------
    cmap : Colormap
        This is a valid matplotlib cm Colormap object that can be used for
        both plotting and colorbar creation.
    norm : Normalize
        This is a valid matplotlib Normalize object that can be used for
        both plotting and colorbar creation.
    cbar_ticks : list
        List of values where tick marks should be placed on colorbar
    cbar_tick_labels : list, optional
        List of tick label values
    boundaries : list, optional
        List of boundaries to use in matplotlib plotting and colorbar
        creation
    cbar_label, optional
        The label for the colorbar
    cbar_spacing : str, default='proportional'
        One of 'proportional' or 'uniform'
    create_colorbar : bool, default=True
        True if colorbar should be created with the set of color info, False otherwise
    cbar_full_width : bool, default=False
        True if colorbar should be full width of figure, center 50% if False

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of mpl_colors_info for use in plotting and colorbar
        creation.
    """
    mpl_colors_info = {}
    mpl_colors_info["cmap"] = cmap
    mpl_colors_info["norm"] = norm
    mpl_colors_info["cbar_ticks"] = cbar_ticks
    mpl_colors_info["cbar_tick_labels"] = cbar_tick_labels
    mpl_colors_info["cbar_label"] = cbar_label
    mpl_colors_info["boundaries"] = boundaries
    mpl_colors_info["cbar_spacing"] = cbar_spacing
    mpl_colors_info["colorbar"] = create_colorbar
    mpl_colors_info["cbar_full_width"] = cbar_full_width
    return mpl_colors_info


def from_ascii(fname, cmap_name=None, reverse=False):
    """Create a ListedColormap instance from an ASCII file of RGB values.

    Parameters
    ----------
    fname : str
        Full path to ascii RGB colortable file
    cmap_name : str, default=None (basename(fname))
        Identifying name of colormap - if None, default to basename(fname)
    reverse : bool, default=False
        If True, reverse the colormap

    Returns
    -------
    cmap : ListedColormap object
        If cmap_name not specified, the colormap name will be the os.path.basename
        of the file.

    Notes
    -----
     * Lines preceded by '#' are ignored.
     * 0-255 or 0-1.0 RGB values (0-255 values are normalized to 0-1.0
       for matplotlib usage)
     * One white space delimited RGB value per line
    """
    # Read data from ascii file into an NLines by 3 float array, skipping
    # lines preceded by "#"
    lines = []
    with open(fname) as palette:
        for line in palette.readlines():
            if line.strip()[0] != "#":
                lines += [line]

    import numpy

    carray = numpy.zeros([len(lines), len(lines[0].strip().split())])
    # carray = numpy.zeros([len(lines), 3])
    for num, line in enumerate(lines):
        carray[num, :] = [float(val) for val in line.strip().split()]

    # Normalize from 0-255 to 0.0-1.0
    if carray.max() > 1.0:
        carray /= 255.0

    # Test to be sure all color array values are between 0.0 and 1.0
    if not (carray.min() >= 0.0 and carray.max() <= 1.0):
        raise ValueError("All values in carray must be between 0.0 and 1.0.")

    if reverse is True:
        carray = numpy.flipud(carray)

    from matplotlib import colors
    from os.path import basename as pathbasename

    if cmap_name is not None:
        cmap_name = pathbasename(fname)
    cmap = colors.ListedColormap(carray, name=cmap_name)
    return cmap


def create_linear_segmented_colormap(
    cmapname, min_val, max_val, transition_vals, transition_colors
):
    """
    Create a LinearSegmentedColormap instance.

    Parameters
    ----------
    cmapname : str
        Name to attach to the matplotlib.color ColorMap object
    min_val : float
        Overall minimum value for the colormap
        Range must be normalized between 0 and 1
    max_val : float
        Overall maximum value for the colormap
        Range must be normalized between 0 and 1
    transition_vals : array-like
        A list of value ranges specified as tuples for generating a
        specific range of colors ie [(0, 10), (10, 30), (30, 60)]
    transition_colors : array-like
        A list of color ranges specified as tuples for generating a
        specific range of colors corresponding to the transition_vals
        (see Notes below)

    Returns
    -------
    cm : LinearSegmentedColormap
        matplotlib colormap object

    Notes
    -----
    Transition colors specified as::

            [('yellow', 'orange'),
             ('pink', 'red'),
             ('violet', 'purple')]

    Where::

        TRANSITIONPOINT1 = 0.0
        TRANSITIONPOINT4 = 1.0
        cmdict = { 'red' :  ((TRANSITIONPOINT1, IGNORED, 1to2STARTCOLOR),
                         (TRANSITIONPOINT2, 1to2ENDCOLOR, 2to3STARTCOLOR),
                         (TRANSITIONPOINT3, 2to3ENDCOLOR, 3to4STARTCOLOR),
                         (TRANSITIONPOINT4, 3to4ENDCOLOR, IGNORED)),
               'green' :  ((TRANSITIONPOINT1, IGNORED, 1to2STARTCOLOR),
                         (TRANSITIONPOINT2, 1to2ENDCOLOR, 2to3STARTCOLOR),
                         (TRANSITIONPOINT3, 2to3ENDCOLOR, 3to4STARTCOLOR),
                         (TRANSITIONPOINT4, 3to4ENDCOLOR, IGNORED)),

               'blue' :  ((TRANSITIONPOINT1, IGNORED, 1to2STARTCOLOR),
                         (TRANSITIONPOINT2, 1to2ENDCOLOR, 2to3STARTCOLOR),
                         (TRANSITIONPOINT3, 2to3ENDCOLOR, 3to4STARTCOLOR),
                         (TRANSITIONPOINT4, 3to4ENDCOLOR, IGNORED)),
            }
    """
    from matplotlib.colors import ColorConverter, LinearSegmentedColormap

    # Sort transitions on start_val
    bluetuple = ()
    greentuple = ()
    redtuple = ()
    start_color = None
    end_color = None
    old_end_color = [0, 0, 0]
    for transition_val, transition_color in zip(transition_vals, transition_colors):
        start_val = transition_val[0]
        end_val = transition_val[1]
        tstart_color = transition_color[0]
        tend_color = transition_color[1]
        # Must start with 0.0 !
        transition_point = (start_val - min_val) / float((max_val - min_val))
        cc = ColorConverter()
        # Convert start/end color to string, tuple, whatever matplotlib can use.
        try:
            start_color = cc.to_rgb(str(tstart_color))
        except ValueError:
            # Allow for tuples as well as string representations
            start_color = cc.to_rgb(ast.literal_eval(str(tstart_color)))
        try:
            end_color = cc.to_rgb(str(tend_color))
        except ValueError:
            end_color = cc.to_rgb(ast.literal_eval(str(tend_color)))
        bluetuple += ((transition_point, old_end_color[2], start_color[2]),)
        redtuple += ((transition_point, old_end_color[0], start_color[0]),)
        greentuple += ((transition_point, old_end_color[1], start_color[1]),)
        LOG.info(
            "    Transition point: "
            + str(transition_point)
            + ": "
            + str(start_val)
            + " to "
            + str(end_val)
        )
        LOG.info(
            "        Start color: %-10s %-40s", str(tstart_color), str(start_color)
        )
        LOG.info("        End color:   %-10s %-40s", str(tend_color), str(end_color))
        old_end_color = end_color
    # Must finish with 1.0 !
    transition_point = (end_val - min_val) / float((max_val - min_val))
    bluetuple += ((transition_point, old_end_color[2], start_color[2]),)
    redtuple += ((transition_point, old_end_color[0], start_color[0]),)
    greentuple += ((transition_point, old_end_color[1], start_color[1]),)

    cmdict = {"red": redtuple, "green": greentuple, "blue": bluetuple}

    cm = LinearSegmentedColormap(cmapname, cmdict)

    return cm



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Code to implement GeoipsCommand Abstract Base Class for the CLI.

Will implement a plethora of commands, but for the meantime, we'll work on
'geoips config','geoips describe', 'geoips list', 'geoips run', 'geoips test', and
'geoips validate'.
"""

import abc
import argparse
from importlib import metadata, resources
import json
from os.path import dirname
from shutil import get_terminal_size

from colorama import Fore, Style
from tabulate import tabulate
import yaml

from geoips.commandline.cmd_instructions import cmd_instructions, alias_mapping
from geoips.commandline.log_setup import setup_logging


class PluginPackages:
    """Class to hold the plugin packages and their paths.

    This class is used to hold the plugin packages and their paths, which are used
    throughout the CLI to determine which plugins are available to the user. This
    class is used in the GeoipsCLI class to determine which plugins are available to
    the user.

    Moving this to a class reduced GeoIPS CLI startup time by about 30%.
    """

    def __init__(self):
        """Initialize the PluginPackages class.

        Initialize the plugin packages and their paths. This is done by using the
        get_plugin_packages() and get_plugin_package_paths() functions.
        """
        self.entrypoints = [
            ep.value
            for ep in sorted(metadata.entry_points(group="geoips.plugin_packages"))
        ]
        self.paths = [
            dirname(resources.files(ep.value))
            for ep in sorted(metadata.entry_points(group="geoips.plugin_packages"))
        ]


plugin_packages = PluginPackages()


class ParentParsers:
    """Object containing shared arguments for commands in a hierarchical order.

    The name of the parsers in this class directly link to the name of the top-level
    command which will share its arguments. For example, since our top-level command is
    'geoips' (see GeoipsCLI), then the arguments created by 'geoips_parser' will be
    shared amongst all of its child commands. As such, every argument created by
    'list_parser' will be shared by 'list' commands, and so on. So long as the name of
    the class matches the structure <class_name>_parser, then the arguments will be
    shared correctly.
    """

    geoips_parser = argparse.ArgumentParser()
    geoips_parser.add_argument(
        "-log",
        "--log-level",
        type=str,
        default="interactive",
        # Specified in order of what will be shown. First is lowest level (10).
        choices=["debug", "info", "warning", "interactive", "error", "critical"],
        help="Log level to output when using the CLI.",
    )

    list_parser = argparse.ArgumentParser(add_help=False)
    list_parser.add_argument(
        "--package_name",
        "-p",
        type=str,
        default="all",
        choices=plugin_packages.entrypoints,
        help="The GeoIPS package to list from.",
    )
    mutex_group = list_parser.add_mutually_exclusive_group()
    mutex_group.add_argument(
        "--long",
        "-l",
        default=True,
        action="store_true",
        help="Flag representing the 'long' listing of a certain command.",
    )
    mutex_group.add_argument(
        "--columns",
        "-c",
        type=str,
        nargs="+",
        default=None,
        help="""Specific Headers of Data you'd like to see listed.
                For more in formation on headers available, run
                'geoips list <cmd> <positional_args> --columns help'.
                """,
    )

    config_parser = argparse.ArgumentParser(add_help=False)
    config_parser.add_argument(
        "-n",
        "--namespace",
        default="geoips.plugin_packages",
        type=str,
        help=(
            "The namespace of plugin packages to create plugin registries for. "
            "If not specified, this defaults to 'geoips.plugin_packages'."
        ),
    )
    config_parser.add_argument(
        "-p",
        "--packages",
        default=None,
        nargs="+",
        type=str,
        help=(
            "The plugin packages to create or delete plugin registries for. Defaults to"
            " None. If None, all plugin packages under 'namespace' will have their "
            "plugin registries created or deleted, based on the command supplied."
        ),
    )


class GeoipsCommand(abc.ABC):
    """Abstract Base Class for top-level GeoIPS Command Classes, such as run or list.

    This class is a blueprint of what each top-level GeoIPS Command Classes should
    implement. Includes shared attributes and an ``add_suparsers`` function which is
    used for initializing command classes of a certain GeoIPS Command.
    """

    def __init__(self, LOG=None, parent=None, legacy=False):
        """Initialize GeoipsCommand with a subparser and default to the command func.

        Do this for each GeoipsCLI.geoips_command_classes. This will instantiate
        each subcommand class with a parser and point towards the correct default
        function to call if that subcommand has been called.

        Parameters
        ----------
        LOG: optional - Logger Object
            - Logging utility which can be used by any command class. Defaults to
              LOG.interactive, however, can be changed to one of the values in
              ["debug", "error", "info", "interactive", "warning"] if
              '--log_level/--log <log_level_name>' is specified at the command line.
        parent: optional - GeoipsCommand Class
            - The parent command class that possibly is initializing it's child.
              Ex. GeoipsList would invoke this init function for each of its subcommand
              classes (GeoipsListPackages, GeoipsListScripts, ...). When it invokes this
              init, it supplies 'self' as an argument to follow the correct logic below.
        legacy: optional - bool
            - Whether or not the command supplied was called in legacy mode. This only
              occurs for 'geoips run' commands, where instead of typing 'geoips run',
              the user called 'run_procflow' or 'data_fusion_procflow'. This is used for
              suppressing or displaying help information for '--procflow'.
        """
        self.legacy = legacy
        # This is the Logger Object, not the actual logger call function of 'log_level'.
        # So, a command class would use the logger via:
        # self.LOG.<log_level>(log_statement)
        self.LOG = LOG
        self.github_org_url = "https://github.com/NRLMMD-GEOIPS/"
        self.parent = parent
        self.alias_mapping = alias_mapping
        if self.parent:
            # Set the combined name of the provided object. For example, if this was the
            # parent 'list' command, it would be 'geoips_list'. If it was a child of
            # list, for example 'scripts', combined name would be 'geoips_list_scripts'
            self.combined_name = f"{parent.combined_name}_{self.name}"

            # The logic below traverses up from the current command class through all
            # of its parents. For example, if this was GeoipsListPackages, the parent
            # traversal would look like GeoipsList -> GeoipsCLI, which would result in
            # parent_parsers looking like [geoips_parser, list_parser]. All of the
            # arguments created by those parent parsers would be inherited by
            # GeoipsListPackages
            curr_parent = self.parent
            self.parent_parsers = []
            while curr_parent and hasattr(ParentParsers, f"{curr_parent.name}_parser"):
                # Don't add parent parser arguments to 'geoips config install'. Config's
                # shared arguments only apply to 'create-registries' and
                # 'delete-registries' and the time being. I've looked for easier
                # workarounds and can't find any suitable options other than hardcoding.

                # If we end up needing to add more hardcoded special cases here, it
                # might be worth looking into other methods of how to share arguments
                # among specific commands, or subsets of commands. For now, this works
                if self.name == "install" and self.parent.name == "config":
                    break
                self.parent_parsers.insert(
                    0,
                    getattr(ParentParsers, f"{curr_parent.name}_parser"),
                )
                curr_parent = curr_parent.parent

            if parent.cmd_instructions:
                # this is used for testing purposes to ensure failure for invalid
                # help information. If the parent already has cmd_instructions set,
                # use these instructions so we can test proper functionality of the CLI.
                self.cmd_instructions = parent.cmd_instructions
            else:
                # Otherwise use the default cmd_instructions which are used for normal
                # invocation of the CLI.
                self.cmd_instructions = cmd_instructions
            try:
                # If the command's name exists w/in the alias mapping, then
                # add thoss aliases to the parser, otherwise just set it as an empty
                # list.
                aliases = self.alias_mapping.get(self.name.replace("_", "-"), [])
                # Attempt to create a sepate sub-parser for the specific command
                # class being initialized so we can separate the commands arguments
                # in a tree-like structure
                self.parser = parent.subparsers.add_parser(
                    self.name,
                    description=self.cmd_instructions["instructions"][
                        self.combined_name
                    ]["help_str"],
                    help=self.cmd_instructions["instructions"][self.combined_name][
                        "help_str"
                    ],
                    usage=self.cmd_instructions["instructions"][self.combined_name][
                        "usage_str"
                    ],
                    parents=self.parent_parsers,
                    conflict_handler="resolve",
                    aliases=aliases,
                    formatter_class=argparse.RawTextHelpFormatter,
                )
            except KeyError:
                raise KeyError(
                    "Error, the supplied command line instructions are improperly "
                    "formatted. You need an 'instructions' entry that contains a "
                    f"'{self.combined_name}' key."
                )
        else:
            # Otherwise initialize a top-level parser for this command.
            self.parser = argparse.ArgumentParser(
                self.name,
                conflict_handler="resolve",
                parents=[ParentParsers.geoips_parser],
                formatter_class=argparse.RawTextHelpFormatter,
            )
            self.LOG = self._get_cli_logger()
            self.combined_name = self.name

        self.add_subparsers()
        self.parser.set_defaults(
            command=self.combined_name.replace("_", " "),
            command_parser=self.parser,
        )

    def _get_cli_logger(self):
        """Set up and retrieve the logger object for use in the CLI.

        If either flag ['--log-level', '--log'] was provided with a valid log level
        after that flag, then set up the logger using the provided log level as the
        filter for what will be logged.

        Log Levels
        ----------
        The log level filters what is logged when the CLI is ran. Filtering order shown
        below. Log levels omit all levels that are below it:
        - interactive
        - debug
        - info
        - warning
        - error
        """
        # An independent parser is needed as this overrides the help messages of the CLI
        # by providing ``add_help=False``, we keep the custom help messages for the CLI
        # and by providing a parent class to the top level CLI parser, then it will be
        # shown in the help messages as well.
        independent_parser = argparse.ArgumentParser(add_help=False)
        independent_parser.add_argument(
            "-log",
            "--log-level",
            type=str,
            default="interactive",
            choices=["interactive", "debug", "info", "warning", "error"],
        )
        # Parse now, as we'll use logging among all of the child command classes
        known_args, remaining_args = independent_parser.parse_known_args()  # NOQA
        log_level = known_args.log_level
        # Set up logging based on the log level provided or defaulted to
        LOG = setup_logging(logging_level=log_level.upper())
        return LOG

    @property
    @abc.abstractmethod
    def name(self):
        """Name of the command class."""
        pass

    @property
    @abc.abstractmethod
    def command_classes(self):
        """List of command_classes related to the top level command.

        For example, if the class provided was GeoipsList, command_classes would
        be the list of available command_classes that "geoips list" implements.
        """
        pass

    def add_subparsers(self):
        """Add subparsers for each command class.

        This is done so we can limit the scope of what arguments are accepted for each
        geoips <cmd> command. This is only done for the top-level command, such as
        "list", "run", "describe", etc.

        For example, if this were the GeoipsList Command Class, we would create a
        self.list_subparsers attribute, which we then add individual parsers for each
        command, as in interfaces, plugins, packages, scripts, etc.
        """
        if len(self.command_classes):
            self.subparsers = self.parser.add_subparsers(
                help=f"{self.name} instructions.",
            )
            for subcmd_cls in self.command_classes:
                subcmd_cls(LOG=self.LOG, parent=self, legacy=self.legacy)

    @property
    def plugin_package_names(self):
        """List of names of all installed Geoips Plugin Packages."""
        return plugin_packages.entrypoints

    @property
    def plugin_package_paths(self):
        """List of paths to all installed Geoips Plugin Packages."""
        return plugin_packages.paths


class GeoipsExecutableCommand(GeoipsCommand):
    """Abstract Base Class for executable CLI commands, inheriting from GeoipsCommand.

    This class is a blueprint of what each executable GeoIPS Command Classes
    can implement.
    """

    def __init__(self, LOG, parent=None, legacy=False):
        """Initialize GeoipsExecutableCommand.

        This is a child of GeoipsCommand and will invoke the functionaly of
        GeoipsCommand __init__ func alongside additional logic needed to set up
        executable-based commands. This will instantiate each subcommand class with a
        parser and point towards the correct default function to call if that subcommand
        has been called.

        Parameters
        ----------
        LOG: Logger Object
            - Logging utility which can be used by any command class. Defaults to
              LOG.interactive, however, can be changed to one of the values in
              ["debug", "error", "info", "interactive", "warning"] if
              '--log_level/--log <log_level_name>' is specified at the command line.
        parent: optional - GeoipsCommand Class
            - The parent command class that possibly is initializing it's child.
              Ex. GeoipsList would invoke this init function for each of its subcommand
              classes (GeoipsListPackages, GeoipsListScripts, ...). When it invokes this
              init, it supplies 'self' as an argument to follow the correct logic below.
        legacy: optional - bool
            - Whether or not the command supplied was called in legacy mode. This only
              occurs for 'geoips run' commands, where instead of typing 'geoips run',
              the user called 'run_procflow' or 'data_fusion_procflow'. This is used for
              suppressing or displaying help information for '--procflow'.
        """
        super().__init__(LOG=LOG, parent=parent, legacy=legacy)
        # Since this class is exectuable (ie. not the cli, top-level list...),
        # add available arguments for that command and set that function to
        # the command's executable function (__call__) if that command is called.
        self.add_arguments()
        self.parser.set_defaults(
            exe_command=self.__call__,
        )

    @property
    def terminal_width(self):
        """The Width in ANSI-Characters of the User's Terminal.

        Generate this every time as the screen width may change during usage.
        """
        return get_terminal_size().columns - 1

    @abc.abstractmethod
    def add_arguments(self):
        """Add arguments related to the command class.

        This is an abstract method because we don't know which arguments need to be
        added for each class at this moment.
        """
        pass

    @abc.abstractmethod
    def __call__(self, args):
        """Functionality to execute the command being called.

        Parameters
        ----------
        args: argparse Namespace()
            - The namespace of the arguments of a specific CLI Command.
        """
        pass

    def _output_dictionary_highlighted(self, dict_entry):
        """Print to terminal the yaml-dumped dictionary of a certain interface/plugin.

        Color the key, value pairs cyan, yellow to highlight the text in a human
        readable manner. This is done for every `geoips describe ...` command.

        Parameters
        ----------
        plugin_entry: dict
            - The dictionary of info for a certain plugin in the plugin registry.
        """
        yaml_text = yaml.dump(dict_entry, default_flow_style=False)
        print()
        for line in yaml_text.split("\n"):
            # Color the keys in cyan and values in yellow
            if ":" in line:
                key, value = line.split(":", 1)
                key = key.title().replace("_", " ")
                if key in ["Package", "Geoips Package"]:
                    key = "GeoIPS Package"
                elif key == "Relpath":
                    key = "Relative Path"
                formatted_line = Fore.CYAN + key + ":" + Style.RESET_ALL
                formatted_line += Fore.YELLOW + value + Style.RESET_ALL
                print(formatted_line)
            else:
                formatted_line = "\t" + Fore.YELLOW + line + Style.RESET_ALL
                print(formatted_line)

    def _get_registry_by_interface_and_package(self, interface, package_name):
        """Retrieve the correct plugin registry given interface and package name.

        Given a GeoIPS Interface and a package name, load in the correct plugin
        registry. This could be the global plugin registry which each interface has
        access to, or it could be the correct interface entry of a certain plugin
        package's "registered_plugins.json", if we are just searching through a
        single package (which occurs when -p <package_name> is given).

        Parameters
        ----------
        interface: geoips.interfaces.<interface_type>
            - The interface we will be parsing and displaying.
        package_name: str
            - The name of the package to retrive the appropriate interfaces from:
              If name is "all", just use the plugin registry that's on all interfaces.

        Returns
        -------
        interface_registry: dict
            - The plugin registry associated with the given interface. If the provided
              interface doesn't exist in a certain package's plugin registry, then None
              is returned instead.
        interface_type: str
            - The type of interface we are dealing with:
              ("module_based", "yaml_based", "text_based")
        """
        if package_name == "all":
            # If there are no plugins of current interface, just return None, do not
            # fail catastrophically. This will fail on "sector_adjusters" interface
            # during "geoips list plugins" if only geoips repo is installed (since there
            # are no "sector_adjuster" plugins in the geoips repo)
            if (
                interface.name
                in interface.plugin_registry.registered_plugins[
                    interface.interface_type
                ]
            ):
                interface_registry = interface.plugin_registry.registered_plugins[
                    interface.interface_type
                ][interface.name]
            else:
                interface_registry = None
        else:
            with open(
                resources.files(package_name) / "registered_plugins.json", "r"
            ) as fo:
                interface_registry = json.load(fo)
            if interface.name in interface_registry[interface.interface_type]:
                interface_registry = interface_registry[interface.interface_type][
                    interface.name
                ]
            else:
                interface_registry = None
        return interface_registry

    def _get_headers_by_command(self, args, default_headers):
        """Retrieve the appropriate headers from a list command with the given args.

        Headers are the top-row of the tabular output from 'list' commands. These are
        customizable via optional args [--column, --long] and we retrieve the
        appropriate headers based on those optional arguments and the 'list' command
        provided.

        Parameters
        ----------
        args: argparse Argument Namespace
            - The arguments provided to a certain list command

        Returns
        -------
        headers: list of str
            - The list of column headers that will be outputted
        """
        columns = args.columns
        if columns:
            headers = {}
            for col in columns:
                if col not in list(default_headers.keys()):
                    err_str = (
                        f"Column header '{col}', is not a valid header. Please select "
                        "one or more of the following keys, which correspond to the "
                        f"appropriate value:\n{default_headers}"
                    )
                    self.parser.error(err_str)
                headers[col] = default_headers[col]
        else:
            # long has been set or was defaulted to; use the default headers
            headers = default_headers
        return headers

    def _print_plugins(self, interface, interface_registry, args):
        """Print the plugins under a certain interface in alongside minimal info.

        "Long Format" includes these pieces of information:
            - GeoIPS Package
            - Interface Name
            - Interface Type
            - Family
            - Plugin Name
            - Source Names
            - Relpath

        Parameters
        ----------
        interface: geoips.interfaces.<interface_type>
            - The interface we will be parsing and displaying.
        interface_registry: dict
            - The plugin registry associated with the given interface.
        args: argparse Argument Namespace
            - The arguments provided to a certain list command
        """
        default_headers = {
            "package": "GeoIPS Package",
            "interface": "Interface Name",
            "plugin_type": "Interface Type",
            "family": "Family",
            "plugin_name": "Plugin Name",
            "source_names": "Source Names",
            "relpath": "Relative Path",
        }
        headers = self._get_headers_by_command(args, default_headers)
        table_data = self._generate_table_data_by_interface(
            interface,
            interface_registry,
            headers,
        )
        print(
            tabulate(
                table_data,
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=self.terminal_width // len(headers),
            )
        )

    def _generate_table_data_by_interface(self, interface, interface_registry, headers):
        """Generate the table data needed for output in '_print_plugins.

        Given a certain interface and a list of headers, generate a list of table data
        used for outputting information used by 'geoips list interface' and
        'geoips list plugins'.

        Parameters
        ----------
        interface: geoips.interfaces.<interface_type>
            - The interface object we will be gathering data from.
        interface_registry: dict
            - The plugin registry associated with the given interface.
        headers: dict
            - Dictionary of strings representing the key-value mapping of the
              headers we'd like to output.

        Returns
        -------
        table_data: 2D List of Strings
            - The corresponding data we will output, where each sub-list is ordered
              by headers.
        """
        table_data = []
        if interface.name == "products":
            for plugin_key in sorted(interface_registry.keys()):
                product_dict = interface_registry[plugin_key]
                table_data += self._get_entry(product_dict, headers)
        else:
            table_data += self._get_entry(interface_registry, headers)
        return table_data

    def _get_entry(self, plugin_dict, headers):
        """Generate a table of info based on the plugin and headers provided.

        Where a table of info is a 2D list of strings that represents the
        plugin x headers combination of data. Say, for product plugin yaml file
        abi.yaml, we'd loop over every product in that file, add to that product
        plugin's entry each header ('source_name', 'family', 'plugin_name', etc.) that
        was selected, and return the 2D table where each 1D list represents the info
        that corresponds to the product plugin.

        Parameters
        ----------
        plugin_dict: dict
            - The portion of the plugin_registry needed to access all plugins of a
              certain interface or all plugins of a certain product.
        headers: dict
            - Dictionary of strings representing the key-value mapping of the
              headers we'd like to output.

        Returns
        -------
        table_data: 2D list of str
            - A 2D list of strings containing information about plugins listed in the
              order of headers.keys()
        """
        table_data = []
        for plugin_key in list(plugin_dict.keys()):
            plugin_entry = []
            for header in list(headers.keys()):
                if (
                    header == "source_names"
                    and plugin_dict[plugin_key]["interface"] != "products"
                ):
                    plugin_entry.append("N/A")
                elif (
                    header == "family"
                    and plugin_dict[plugin_key]["interface"] == "products"
                ):
                    plugin_entry.append("N/A")
                elif header == "plugin_name":
                    plugin_entry.append(plugin_key)
                else:
                    plugin_entry.append(plugin_dict[plugin_key][header])
            table_data.append(plugin_entry)
        return table_data


class CommandClassFactory:
    """An abstract class factory for creating CLI-based Command Classes.

    This architecture can be used to generate CLI command classes for specific use
    cases. A couple which we'll implement early on is are:

    * GeoipsListSingleInterface:

        * GeoipsListSingleInterfaceAlgorithms
        * GeoipsListSingleInterfaceColormappers
        * ...
        * GeoipsListSingleInterfaceTitleFormatters

    * GeoipsDescribeArtifact

        * GeoipsDescribeArtifactAlgorithm
        * GeoipsDescribeArtifactColormapper
        * ...
        * GeoipsDescribeArtifactTitleFormatter

    This class has been created to reduce the verbosity of geoips commands without
    having to copy-paste classes specifc to a certain interface.
    """

    def __init__(self, base_class, class_name, add_attrs={}):
        """Initialize the class factory and generate a class derived from base_class.

        Parameters
        ----------
        base_class: Class
            - The base class to build each generated class with..
            - Ie. if we were creating a class factory for GeoipsListSingleInterface, the
              base class would be 'GeoipsListSingleInterface'.
        class_name: str
            - The name of the class that we'd like to generate.
            - Ie. if we were creating a class factory for GeoipsListSingleInterface, the
              class to generate would be a strign that represents all the class for the
              associated interface we'd like to generate ('algorithms', ...).
            - The classes would then look like:
                - "GeoipsListAlgorithms"
                - "GeoipsListColormappers"
                - ...
                - "<base_class_name><class_name>"
        add_attrs: dict (optional)
            - A dictionary of attributes we'd like to assign to each command class
            - If specified and has overlapping keys to those specified below, this
              dictionary will override the defaults.
        """
        self.base_class = base_class
        self.base_class_name = self.base_class.__name__
        default_attrs = {
            "name": class_name,
            "command_classes": [],
            "add_arguments": base_class.add_arguments,
            "__call__": base_class.__call__,
        }
        # Combine with additional attributes, overwriting defaults where applicable.
        class_attrs = {**default_attrs, **add_attrs}
        # Add the class of <base_class_name><class_name> to the constructors 'classes'
        # attribute. Don't actually initialize these classes as we'll be doing that
        # later once we have the required information.
        self.generated_class = type(
            f"{self.base_class_name}{class_name.title().replace('_', '')}",
            (base_class,),
            class_attrs,
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geoips plugin for driving pyresample Nearest Neighbor interpolation."""

import logging

import xarray

from geoips.plugins.modules.interpolators.utils.interp_pyresample import (
    interp_kd_tree,
    get_data_box_definition,
)
from geoips.data_manipulations.info import percent_unmasked
from geoips.geoips_utils import copy_standard_metadata, remove_unsupported_kwargs

LOG = logging.getLogger(__name__)

interface = "interpolators"
family = "2d"
name = "interp_nearest"


def get_final_roi(xarray_obj, area_def):
    """Get the final interpolation Radius of Influence.

    This takes the maximum of the xarray attribute, area_def pixel width,
    and area_def pixel height.
    """
    roi = None
    if "interpolation_radius_of_influence" in xarray_obj.attrs.keys():
        roi = xarray_obj.interpolation_radius_of_influence
        LOG.info("    USING XARRAY ROI of %s", roi)
    if not roi or area_def.pixel_size_x > roi:
        roi = area_def.pixel_size_x
        LOG.info("    USING AREA_DEF X ROI of %s", roi)
    if area_def.pixel_size_y > roi:
        roi = area_def.pixel_size_y
        LOG.info("    USING AREA_DEF Y ROI of %s", roi)
    return roi


def call(area_def, input_xarray, output_xarray, varlist, array_num=None, **kwargs):
    """Pyresample interp_kd_tree nearest neighbor GeoIPS plugin."""
    LOG.info(
        "Interpolating nearest using standard scifile register method: kd_tree nearest"
    )

    roi = get_final_roi(input_xarray, area_def)

    vars_to_interp = []
    if array_num is not None:
        if len(input_xarray[varlist[0]].shape) == 2:
            if len(input_xarray["latitude"].shape) == 2:
                lons = input_xarray["longitude"].to_masked_array()[:, array_num]
                lats = input_xarray["latitude"].to_masked_array()[:, array_num]
            else:
                lons = input_xarray["longitude"].to_masked_array()
                lats = input_xarray["latitude"].to_masked_array()
            for varname in varlist:
                vars_to_interp += [
                    input_xarray[varname].to_masked_array()[:, array_num]
                ]
        else:
            if len(input_xarray["latitude"].shape) == 2:
                lons = input_xarray["longitude"].to_masked_array()
                lats = input_xarray["latitude"].to_masked_array()
            else:
                lons = input_xarray["longitude"].to_masked_array()[:, :, array_num]
                lats = input_xarray["latitude"].to_masked_array()[:, :, array_num]
            for varname in varlist:
                vars_to_interp += [
                    input_xarray[varname].to_masked_array()[:, :, array_num]
                ]
    else:
        lons = input_xarray["longitude"].to_masked_array()
        lats = input_xarray["latitude"].to_masked_array()
        for varname in varlist:
            vars_to_interp += [input_xarray[varname].to_masked_array()]

    # Use standard scifile / pyresample registration
    data_box_definition = get_data_box_definition(input_xarray.source_name, lons, lats)

    kd_kwargs = remove_unsupported_kwargs(interp_kd_tree, kwargs)
    interp_data = interp_kd_tree(
        vars_to_interp,
        area_def,
        data_box_definition,
        float(roi),
        interp_type="nearest",
        **kd_kwargs,
    )

    for arr, orig, varname in zip(interp_data, vars_to_interp, varlist):
        LOG.info("%s min/max before: %s to %s", varname, orig.min(), orig.max())
        LOG.info("%s min/max after:  %s to %s", varname, arr.min(), arr.max())
        LOG.info("%s Percent unmasked before %s", varname, percent_unmasked(orig))
        LOG.info("%s Percent unmasked after  %s", varname, percent_unmasked(arr))

    if output_xarray is None:
        output_xarray = xarray.Dataset()
    if "latitude" not in output_xarray.variables.keys():
        interp_lons, interp_lats = area_def.get_lonlats()
        output_xarray["latitude"] = xarray.DataArray(interp_lats)
        output_xarray["longitude"] = xarray.DataArray(interp_lons)
    copy_standard_metadata(input_xarray, output_xarray, force=False)
    output_xarray.attrs["registered_dataset"] = True
    output_xarray.attrs["area_definition"] = area_def

    for ind in range(len(varlist)):
        output_xarray[varlist[ind]] = xarray.DataArray(interp_data[ind])

    return output_xarray



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `describe interface` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from geoips import interfaces
from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsDescribeInterface(BaseCliTest):
    """Unit Testing Class for Describe Interface Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the this command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "describe"]
            # add arguments for retrieving each GeoIPS Interface
            for interface_name in interfaces.__all__:
                interface_name = interface_name.replace("_", "-")
                for alias in self.alias_mapping[interface_name] + [interface_name]:
                    self._cmd_list.append(base_args + [alias])
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non_existent_interface
            self._cmd_list.append(base_args + ["non_existent_interface"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips describe interface ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # An error occurred using args. Assert that args is not valid and check the
        # output of the error.
        err_str = "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
        assert err_str in error

    def check_output(self, args, output):
        """Ensure that the 'geoips describe interface ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            usg_str = (
                "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
            )
            assert usg_str in output
        else:
            # Checking that output from geoips describe package command is valid
            expected_outputs = [
                "Absolute Path",
                "Docstring",
                "Interface",
                "Interface Type",
                "Supported Families",
            ]
            for output_item in expected_outputs:
                assert f"{output_item}:" in output


test_sub_cmd = TestGeoipsDescribeInterface()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips describe interface ...' commands.

    This test covers every valid combination of commands for the 'geoips describe
    interface' command. We also test invalid commands, to ensure that the proper help
    documentation is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (I.e.
          ['geoips', 'describe', 'interface'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test YAML schema."""
import pytest

import re
import yaml
from jsonschema.exceptions import ValidationError  # noqa
from pathlib import Path

from geoips.interfaces.base import YamlPluginValidator


test_path = Path(__file__).parent
validator = YamlPluginValidator()


def get_error(name):
    """Get error."""
    try:
        exc = getattr(__builtins__, name)
    except AttributeError:
        try:
            exc = globals()[name]
        except KeyError:
            raise ValueError(f"{name} is not a known exception")
    if not issubclass(exc, Exception):
        raise ValueError(f"{name} is not a valid exception")
    return exc


def path_to_id(fname):
    """Get path to id."""
    fpath = Path(fname).relative_to(test_path)
    parts = list(fpath.parent.parts[1:])
    parts.append(fpath.stem)
    id = ".".join(parts)
    return id


def yield_bad_plugins():
    """Yield bad plugins."""
    bad_files = {path_to_id(fname): fname for fname in test_path.rglob("bad/**/*.yaml")}
    for validator_name, bad_file in bad_files.items():
        plugins = yaml.safe_load_all(open(bad_file, "r"))
        for plugin in plugins:
            yield validator_name, plugin


def yield_good_plugins():
    """Yield good plugins."""
    good_files = {
        path_to_id(fname): fname for fname in test_path.rglob("good/**/*.yaml")
    }
    for validator_name, good_file in good_files.items():
        plugins = yaml.safe_load_all(open(good_file, "r"))
        for plugin in plugins:
            yield validator_name, plugin


# plugins_to_test = [
#     "feature_annotators",
#     "gridline_annotators",
#     "product_defaults",
# ]
# good_yaml_plugins = []
# for plg in plugins_to_test:
#     good_yaml_plugins.extend(Path(f"{package_path}/plugins/{plg}").rglob("*.yaml"))


@pytest.mark.parametrize("validator_name, plugin", yield_good_plugins())
def test_is_plugin_valid(validator_name, plugin):
    """Test whether a plugin validates against its interface's schema."""
    validator.validate(plugin, validator_id=validator_name)


@pytest.mark.parametrize("validator_name, plugin", yield_bad_plugins())
def test_is_plugin_invalid(validator_name, plugin):
    """Test that a plugin fails to validate against its interface's schema."""
    try:
        error_pattern = plugin.pop("error_pattern", None)
    except (AttributeError, TypeError):
        error_pattern = None
    try:
        error_name = plugin.pop("error", "ValidationError")
    except (AttributeError, TypeError):
        error_name = "ValidationError"
    error = get_error(error_name)

    # If interface and family are in the plugin, don't pass validator_name
    try:
        plugin["interface"]
        plugin["family"]
        validator_id = None
    except (KeyError, TypeError):
        validator_id = validator_name

    with pytest.raises(error) as excinfo:
        validator.validate(plugin, validator_id=validator_id)
    if error_pattern is not None:
        assert re.match(error_pattern, str(excinfo.value))



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list packages` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListPackages(BaseCliTest):
    """Unit Testing Class for List Packages Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListPackages command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "list", "packages"]
            alias_args = ["geoips", "ls", "packages"]
            self._cmd_list = [base_args, alias_args]
            # Add argument list invoking the --columns flag
            self._cmd_list.append(base_args + ["--columns", "package", "package_path"])
            self._cmd_list.append(alias_args + ["--columns", "package", "package_path"])
            self._cmd_list.append(base_args + ["--columns", "docstring", "package"])
            self._cmd_list.append(alias_args + ["--columns", "docstring", "package"])
            self._cmd_list.append(
                base_args + ["--columns", "docstring", "package_path"]
            )
            self._cmd_list.append(
                alias_args + ["--columns", "docstring", "package_path"]
            )
            self._cmd_list.append(
                base_args + ["--columns", "package", "docstring", "package_path"]
            )
            self._cmd_list.append(
                alias_args + ["--columns", "package", "docstring", "package_path"]
            )
            # Add argument list which invokes the help message for this command
            self._cmd_list.append(base_args + ["-h"])
            self._cmd_list.append(alias_args + ["-h"])
            # Add argument list with a non-existent command call ("-p")
            self._cmd_list.append(base_args + ["-p", "geoips"])
            self._cmd_list.append(alias_args + ["-p", "geoips"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips list packages ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # bad command has been provided, check the contents of the error message
        assert args != ["geoips", "list", "packages"] and args != [
            "geoips",
            "ls",
            "packages",
        ]
        assert "usage: To use, type `geoips list packages`" in error
        assert "Error: '-p' flag is not supported for this command" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips list packages ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        if "usage: To use, type" in output:
            # -h has been called, check help message contents for this command
            assert "-h" in args
            assert "type `geoips list packages`" in output
        else:
            # The args provided are valid, so test that the output is actually correct
            for arg in ["geoips", "list", "packages"]:
                if arg == "list":
                    assert arg in args or "ls" in args
                else:
                    assert arg in args
            # Assert that the correct headers exist in the CLI output
            headers = {
                "GeoIPS Package": "package",
                "Docstring": "docstring",
                "Package Path": "package_path",
                "Version Number": "version",
            }
            selected_cols = self.retrieve_selected_columns(args)
            self.assert_correct_headers_in_output(output, headers, selected_cols)
            # Assert that we found every installed package
            for pkg_name in self.plugin_package_names:
                if "--columns" not in args or "package" in selected_cols:
                    assert pkg_name in output


test_sub_cmd = TestGeoipsListPackages()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list packages ...' commands.

    This test covers every valid combination of commands for the 'geoips list packages'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'list', 'packages'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing Legacy colormap for ~89GHz PMW products."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "pmw_89H_Legacy"


def call(data_range=[180.0, 280.0], cbar_label="TB (K)"):
    """Legacy Colormap for displaying ~89GHz PMW data.

    Parameters
    ----------
    data_range : list of float, default=[180, 280]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include 180 and 280

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    min_tb = data_range[0]
    max_tb = data_range[1]

    if min_tb > 180 or max_tb < 254:
        raise ValueError("89H Legacy TB range must include 180 and 254")

    # use the TeraScan TC 89 GHz legacy color table for 89 GHz products
    # (plus one Black for TB<=180)
    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [(min_tb, 180), (180, 212), (212, 228), (228, 254), (254, max_tb)]
    transition_colors = [
        ("black", "black"),
        ("#A4641A", "#FC0603"),
        ("#F4CD03", "#F2F403"),
        ("#8CF303", "#0FB503"),
        ("#06DCFD", "#0708B5"),
    ]

    # ticks = [xx[0] for xx in transition_vals]

    # special selection of label

    ticks = [180, 190, 200, 210, 220, 230, 240, 250, 260, 270, 280]

    # selection of min and max values for colormap if needed
    min_tb = transition_vals[0][0]
    max_tb = transition_vals[-1][1]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "pmw_89h_legacy", min_tb, max_tb, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_tb, vmax=max_tb)

    cbar_label = "TB (K)"

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_tb, max_tb
    return mpl_colors_info
